"""
Implementation of the ta2ta3 API v2 (preprocessing extensions) -- core.proto
"""

__author__ = "Saswati Ray"
__email__ = "sray@cs.cmu.edu"

import logging
import os, sys, copy, threading
import pandas as pd
import pickle
import numpy as np

from multiprocessing import Pool, cpu_count
from multiprocessing.context import TimeoutError
import uuid
from timeit import default_timer as timer

logging.basicConfig(level=logging.ERROR)
pd.set_option('display.max_rows', None)

from d3m.container.dataset import D3MDatasetLoader, Dataset
from d3m.metadata import base as metadata_base
from d3m import container

# AutonML imports
import autonml.primitive_lib as primitive_lib
import autonml.search as search
import autonml.core_pb2 as core_pb2
import autonml.core_pb2_grpc as core_pb2_grpc
import autonml.value_pb2 as value_pb2
import autonml.primitive_pb2 as primitive_pb2
import autonml.problem_pb2 as problem_pb2
import autonml.pipeline_pb2 as pipeline_pb2

import autonml.solutiondescription as solutiondescription
import autonml.helper_functions as helper_functions 
import autonml.util as util
import autonml.auto_solutions as auto_solutions

class Core(core_pb2_grpc.CoreServicer):
    def __init__(self):
        self._solution_to_search = {}
        self._primitives = {}
        self._solutions = {}
        self._solution_score_map = {}
        self._search_solutions = {}
        self.timeouts = {}
        self.lock = threading.Lock()
        self.async_message_thread = Pool(cpu_count()) #pool.ThreadPool(processes=1,)
        self._primitives = primitive_lib.load_primitives()         
        outputDir = os.environ['D3MOUTPUTDIR']

    def get_task_list(self, keywords):
        names = []
        for k in keywords:
            name = k #problem_pb2.TaskKeyword.Name(k)
            name = name.replace('_', '')
            names.append(name)
        return names

    def get_task_name(self, keywords):
        names = self.get_task_list(keywords)
        return util.get_task(names)

    def search_solutions(self, request, dataset):
        """
        Populate potential solutions for TA3
        """
        primitives = self._primitives
        problem = request.problem.problem
        template = request.template
        task_name = self.get_task_name(problem.task_keywords)
        logging.critical(task_name)

        solutions = []
        pipeline_placeholder_present = False
        basic_sol = None

        # TA3 has specified a pipeline
        if template != None and isinstance(template, pipeline_pb2.PipelineDescription) and len(template.steps) > 0:
            basic_sol = solutiondescription.SolutionDescription(request.problem)
            basic_sol.create_from_pipelinedescription(self._solutions, pipeline_description=template)
            if basic_sol.contains_placeholder() == False:  # Fully defined
                solutions.append(basic_sol)
                return (solutions, 0)
            else: # Placeholder present
                pipeline_placeholder_present = True    
                inputs = []
                inputs.append(dataset)
                new_dataset = basic_sol.fit(inputs=inputs, solution_dict=self._solutions)      
                dataset = new_dataset      
                logging.critical("New datset from specified pipeline: %s", new_dataset)

        taskname = task_name.replace('_', '')
        logging.critical("taskname = %s", taskname)
        metric = request.problem.problem.performance_metrics[0].metric
        posLabel = request.problem.problem.performance_metrics[0].pos_label
        start = timer()
        automl = auto_solutions.auto_solutions(taskname, request.problem)
        solutions = automl.get_solutions(dataset, metric, request.time_bound_search)
        end = timer()
        time_used = end - start
        logging.critical("No. of solutions = %d", len(solutions))
        if pipeline_placeholder_present is True:
            new_solution_set = []
            for s in solutions:
                try:
                    pipe = copy.deepcopy(basic_sol)
                    pipe.id = str(uuid.uuid4())
                    pipe.add_subpipeline(s)
                    self._solutions[s.id] = s 
                    new_solution_set.append(pipe)
                except:
                    logging.critical(sys.exc_info()[0])            
            logging.critical("%s", new_solution_set)
            return (new_solution_set, time_used)
 
        return (solutions, time_used)
    
    def SearchSolutions(self, request, context):
        """
        TA2-3 API call
        """
        logging.critical("Message received: SearchSolutions")
        search_id_str = str(uuid.uuid4())

        self._solution_score_map[search_id_str] = request

        outputDir = os.environ['D3MOUTPUTDIR'] + "/" + search_id_str
        util.initialize_for_search(outputDir)
        return core_pb2.SearchSolutionsResponse(search_id = search_id_str)

    def _get_inputs(self, problem, rinputs):
        inputs = []
 
        for ip in rinputs:
            dataset = None
            if ip.HasField("dataset_uri") == True:
                dataset = D3MDatasetLoader().load(ip.dataset_uri)
            elif ip.HasField("csv_uri") == True:
                data = pd.read_csv(ip.csv_uri, dtype=str, header=0, na_filter=False, encoding='utf8', low_memory=False,)
                dataset = container.DataFrame(data)

            logging.critical("Problem %s", problem.problem)
            if len(problem.inputs) > 0:
                logging.critical("Problem inputs %s", problem.inputs)
                targets = problem.inputs[0].targets
                dataset = util.add_target_metadata(dataset, targets)
                dataset = util.add_privileged_metadata(dataset, problem.inputs[0].privileged_data)
            inputs.append(dataset) 

        return inputs
       
    def GetSearchSolutionsResults(self, request, context):
        """
        TA2-3 API call
        """
        logging.critical("Message received: GetSearchSolutionsRequest")
        search_id_str = request.search_id
 
        start=helper_functions.compute_timestamp()
        msg = core_pb2.Progress(state=core_pb2.PENDING, status="", start=start, end=helper_functions.compute_timestamp())
        yield core_pb2.GetSearchSolutionsResultsResponse(progress=msg, done_ticks=0, all_ticks=0, solution_id="",
                     internal_score=0.0, scores=[])

        request_params = self._solution_score_map[search_id_str]
        count = 0
        self.timeouts[search_id_str] = 1      
        inputs = self._get_inputs(request_params.problem, request_params.inputs)
        (solutions, time_used) = self.search_solutions(request_params, inputs[0])
        self._search_solutions[search_id_str] = []

        # Fully specified solution
        if request_params.template != None and isinstance(request_params.template, pipeline_pb2.PipelineDescription) \
            and len(request_params.template.steps) > 0 and len(solutions) == 1:
            msg = core_pb2.Progress(state=core_pb2.COMPLETED, status="", start=start, end=helper_functions.compute_timestamp())
            count = count + 1
            id = solutions[0].id
            self._solutions[id] = solutions[0]
            self._search_solutions[search_id_str].append(id)
            self._solution_to_search[id] = search_id_str 
            yield core_pb2.GetSearchSolutionsResultsResponse(progress=msg, done_ticks=1, all_ticks=1,
                          solution_id=id, internal_score=0.0, scores=[])
                        
        else: # Evaluate potential solutions
            index = 0
            msg = core_pb2.Progress(state=core_pb2.RUNNING, status="", start=start, end=helper_functions.compute_timestamp())

            metric = request_params.problem.problem.performance_metrics[0].metric
            posLabel = request_params.problem.problem.performance_metrics[0].pos_label
            logging.critical("Search timeout = %d min", request_params.time_bound_search)

            # Check if search has already timed out!!!
            search_for_solutions = True
            self.lock.acquire(timeout=1)
            if search_id_str not in self.timeouts or self.timeouts[search_id_str] == 0:  
                search_for_solutions = False
            self.lock.release() 
 
            if search_for_solutions == False:
                return
         
            results =  [self.async_message_thread.apply_async(search.evaluate_solution_score, (inputs, sol, self._primitives, metric, posLabel, self._solutions, )) for sol in solutions]
            
            # Timeout in min -> timeout in seconods
            timeout = request_params.time_bound_search * 60
            if timeout <= 0:
                timeout = None
            elif timeout > 60:
                timeout = timeout - 45
                timeout = timeout - time_used
                if timeout <= 0:
                    timeout = 1

            if timeout is not None:
                logging.critical("Timeout = %d sec", timeout)

            outputDir = os.environ['D3MOUTPUTDIR'] + "/" + search_id_str
            valid_solution_scores = {}
            valid_indices = []

            # Evaluate potential solutions asynchronously and get end-result
            for r in results:
                try:
                    # Check if search has timed out! OR StopSearch invoked?
                    self.lock.acquire(timeout=1)
                    if search_id_str not in self.timeouts or self.timeouts[search_id_str] == 0:
                        timeout = 0
                    self.lock.release()

                    start_solution = timer()

                    (score, optimal_params) = r.get(timeout=timeout)

                    if timeout == 0:
                        continue

                    count = count + 1
                    id = solutions[index].id
                    valid_indices.append(index)
                    self._search_solutions[search_id_str].append(id)
                    self._solution_to_search[id] = search_id_str
                    valid_solution_scores[index] = score
                    if optimal_params is not None and len(optimal_params) > 0:
                        solutions[index].set_hyperparams(self._solutions, optimal_params)
                    util.write_pipeline_json(solutions[index], self._primitives, self._solutions, outputDir + "/pipelines_searched", outputDir + "/subpipelines")
                    end_solution = timer()
                    time_used = end_solution - start_solution
                    timeout = timeout - time_used
                    if timeout <= 0:
                        timeout = 1
                    sscore = core_pb2.Score(metric=problem_pb2.ProblemPerformanceMetric(metric=metric), value=value_pb2.Value(raw=value_pb2.ValueRaw(double=score)))
                    search_score = core_pb2.SolutionSearchScore(scoring_configuration=core_pb2.ScoringConfiguration(), scores=[sscore])
                    yield core_pb2.GetSearchSolutionsResultsResponse(progress=msg, done_ticks=count, all_ticks=len(solutions), solution_id=id,
                                        internal_score=0.0, scores=[search_score])
                except TimeoutError:
                    # Solution timed out! 
                    logging.critical(solutions[index].primitives)
                    logging.critical(sys.exc_info()[0])
                    logging.critical("Solution terminated: %s", solutions[index].id)
                    timeout = 1
                except:
                    # Error in scoring solution!
                    logging.critical(solutions[index].primitives)
                    logging.critical(sys.exc_info()[0])
                    logging.critical("Solution terminated: %s", solutions[index].id)
                index = index + 1

            for index in valid_indices:
                id = solutions[index].id
                self._solutions[id] = solutions[index]

            # Sort solutions by their scores and rank them
            sorted_x = search.rank_solutions(valid_solution_scores, metric)
            sol_rank = 1
            for (index, score) in sorted_x:
                id = solutions[index].id
                self._solutions[id].rank = sol_rank
                logging.critical("Rank %d", sol_rank)
                print("Score ", score)
                rank = core_pb2.Score(metric=problem_pb2.ProblemPerformanceMetric(metric="RANK"), value=value_pb2.Value(raw=value_pb2.ValueRaw(double=sol_rank)))
                search_rank = core_pb2.SolutionSearchScore(scoring_configuration=core_pb2.ScoringConfiguration(), scores=[rank])
                sscore = core_pb2.Score(metric=problem_pb2.ProblemPerformanceMetric(metric=metric), value=value_pb2.Value(raw=value_pb2.ValueRaw(double=score)))
                search_score = core_pb2.SolutionSearchScore(scoring_configuration=core_pb2.ScoringConfiguration(), scores=[sscore])
                yield core_pb2.GetSearchSolutionsResultsResponse(progress=msg, done_ticks=count, all_ticks=len(solutions), solution_id=id,
                                        internal_score=0.0, scores=[search_rank, search_score])
                sol_rank = sol_rank + 1  
            msg = core_pb2.Progress(state=core_pb2.COMPLETED, status="", start=start, end=helper_functions.compute_timestamp()) 
            yield core_pb2.GetSearchSolutionsResultsResponse(progress=msg, done_ticks=count, all_ticks=count, solution_id="", internal_score=0.0, scores=[])
        
        self._solution_score_map.pop(search_id_str, None)
        logging.critical("No. of sol = %d", count)

    def EndSearchSolutions(self, request, context):
        """
        TA2-3 API call
        """
        logging.critical("Message received: EndSearchSolutions")
        search_id_str = request.search_id

        if search_id_str in self._search_solutions:
            for sol_id in self._search_solutions[search_id_str]:
                self._solutions.pop(sol_id, None)
                self._solution_to_search.pop(sol_id, None)
            self._search_solutions.pop(search_id_str, None)
            self.lock.acquire(timeout=1)
            self.timeouts.pop(search_id_str, None)
            self.lock.release()

        return core_pb2.EndSearchSolutionsResponse()

    def StopSearchSolutions(self, request, context):
        """
        TA2-3 API call
        """
        search_id_str = request.search_id
        logging.critical("Message received: StopSearchSolutions")
        if search_id_str in self.timeouts:
            self.lock.acquire(timeout=1)
            self.timeouts[search_id_str] = 0
            self.lock.release()

        return core_pb2.StopSearchSolutionsResponse()

    def SaveSolution(self, request, context):

        logging.critical("Message received: SaveSolution")

        # Get the solution ID from the request object
        solution_id = request.solution_id

        logging.critical("Solution id:", solution_id)

        # Get the output directory env variable
        output_dir = os.environ['D3MOUTPUTDIR']

        logging.critical("D3MOUTPUTDIR", output_dir)

        if solution_id in self._solutions:

            logging.critical("Solution found.")

            solution_object = {
                'solution_id': solution_id,
                'solution': self._solutions[solution_id]
            }

            logging.critical(solution_object)

            with open(os.path.join(output_dir, '/temp/', solution_id+'.pickle'), 'wb') as f:

                logging.critical("File opened")

                # Pickle the solution to the file
                pickle.dump(solution_object, f, pickle.HIGHEST_PROTOCOL)

                logging.critical("Pickle dumped to file")

                # Get real path of the file
                filepath = os.path.realpath(f.name)

                logging.critical("Real file path:", filepath)

                yield core_pb2.SaveSolutionResponse(solution_uri=filepath)

                logging.critical("Yielded", core_pb2.SaveSolutionResponse(solution_uri=filepath))

        else:

            logging.critical("SaveSolution: Solution %s not found!", solution_id)
            yield core_pb2.SaveSolutionResponse(solution_uri="")

    def LoadSolution(self, request, context):

        logging.critical("Message received: LoadSolution")

        filepath = request.solution_uri

        try:

            with open(filepath, 'rb') as f:

                # Unpickle from file
                solution_object = pickle.load(f)

                # Cache the solution in memory
                self._solutions[solution_object['solution_id']] = solution_object['solution']

                yield core_pb2.LoadSolutionResponse(solution_id=solution_object['solution_id'])

        except:

            logging.critical("LoadSolution: File %s not found!", filepath)
            yield core_pb2.LoadSolutionResponse(solution_id="")

    def SaveFittedSolution(self, request, context):

        logging.critical("Message received: SaveFittedSolution")

        # Get the solution ID from the request object
        solution_id = request.solution_id

        # Get the output directory env variable
        output_dir = os.environ['D3MOUTPUTDIR']

        if solution_id in self._solutions:

            solution_object = {
                'solution_id': solution_id,
                'solution': self._solutions[solution_id]
            }

            with open(os.path.join(output_dir, '/temp/', solution_id + '.pickle'), 'wb') as f:

                # Pickle the solution to the file
                pickle.dump(solution_object, f, pickle.HIGHEST_PROTOCOL)

                # Get real path of the file
                filepath = os.path.realpath(f.name)

                yield core_pb2.SaveFittedSolutionResponse(solution_uri=filepath)

        else:

            logging.critical("SaveFittedSolution: Solution %s not found!", solution_id)
            yield core_pb2.SaveFittedSolutionResponse(solution_uri="")

    def LoadFittedSolution(self, request, context):

        logging.critical("Message received: LoadFittedSolution")

        filepath = request.fitted_solution_uri

        try:

            with open(filepath, 'rb') as f:

                # Unpickle from file
                solution_object = pickle.load(f)

                # Cache the solution in memory
                self._solutions[solution_object['solution_id']] = solution_object['solution']

                yield core_pb2.LoadSolutionResponse(fitted_solution_id=solution_object['solution_id'])

        except:

            logging.critical("LoadFittedSolution: File %s not found!", filepath)
            yield core_pb2.LoadSolutionResponse(fitted_solution_id="")

    def SplitData(self, request, context):
        logging.critical("Message received: SplitData")
        pass

    def ScorePredictions(self, request, context):
        logging.critical("Message received: ScorePredictions")
        pass

    def GetStepDescriptions(self, solution_id):
        param_map = []
        solution = self._solutions[solution_id]
        num_steps = solution.num_steps()
        for j in range(num_steps):
            if solution.primitives[j] is not None:
                step = core_pb2.StepDescription(primitive=core_pb2.PrimitiveStepDescription(hyperparams=solution.get_hyperparams(j, self._primitives)))
            else:
                step_array = self.GetStepDescriptions(solution.subpipelines[j])
                step = core_pb2.StepDescription(pipeline=core_pb2.SubpipelineStepDescription(steps=step_array))
            param_map.append(step)
        return param_map

    def DescribeSolution(self, request, context):
        """
        TA2-3 API call
        """
        logging.critical("Message received: DescribeSolution")
        solution_id = request.solution_id
        solution = self._solutions[solution_id]
        desc = solution.describe_solution(self._primitives, self._solutions)
        param_map = self.GetStepDescriptions(solution_id)

        return core_pb2.DescribeSolutionResponse(pipeline=desc, steps=param_map)

    def ScoreSolution(self, request, context):
        """
        TA2-3 API call
        """
        logging.critical("Message received: ScoreSolution")

        request_id = str(uuid.uuid4())
        self._solution_score_map[request_id] = request

        return core_pb2.ScoreSolutionResponse(request_id = request_id)

    def GetScoreSolutionResults(self, request, context):
        """
        TA2-3 API call
        """
        logging.critical("Message received: GetScoreSolutionResults")
        request_id = request.request_id
        request_params = self._solution_score_map[request_id]
        
        start=helper_functions.compute_timestamp()
        solution_id = request_params.solution_id
        msg = core_pb2.Progress(state=core_pb2.RUNNING, status="", start=start, end=helper_functions.compute_timestamp())
        
        send_scores = []
        from timeit import default_timer as timer

        if solution_id not in self._solutions:
            logging.critical("GetScoreSolutionResults: Solution %s not found!", solution_id)  
            msg = core_pb2.Progress(state=core_pb2.ERRORED, status="", start=start, end=helper_functions.compute_timestamp())
            # Clean up
            self._solution_score_map.pop(request_id, None)
            yield core_pb2.GetScoreSolutionResultsResponse(progress=msg, scores=[])
        else:
            inputs = self._get_inputs(self._solutions[solution_id].problem, request_params.inputs)
            score = 0.0
            try:
                s = timer()                
                (score, optimal_params) = self._solutions[solution_id].score_solution(inputs=inputs, metric=request_params.performance_metrics[0].metric,
                                posLabel = request_params.performance_metrics[0].pos_label,
                                primitive_dict=self._primitives, solution_dict=self._solutions)
                if optimal_params is not None and len(optimal_params) > 0:
                    self._solutions[solution_id].set_hyperparams(self._solutions, optimal_params)

                e = timer()
                logging.critical("Time taken = %s sec", e-s) 
            except:
                score = 0.0
                logging.critical("Exception in score: %s", self._solutions[solution_id].primitives)
                logging.critical("Exception in score: %s", sys.exc_info()[0])
            
            search_id_str = self._solution_to_search[solution_id]
            outputDir = os.environ['D3MOUTPUTDIR'] + "/" + search_id_str
            try:
                util.write_pipeline_json(self._solutions[solution_id], self._primitives, self._solutions, outputDir + "/pipelines_scored", outputDir + "/subpipelines")
            except:
                logging.critical(sys.exc_info()[0])
                logging.critical(self._solutions[solution_id].primitives)
            logging.critical("Score = %f", score)
            send_scores.append(core_pb2.Score(metric=request_params.performance_metrics[0],
             fold=0, value=value_pb2.Value(raw=value_pb2.ValueRaw(double=score)), random_seed=0))

            yield core_pb2.GetScoreSolutionResultsResponse(progress=msg, scores=[]) 

            # Clean up
            self._solution_score_map.pop(request_id, None)

            msg = core_pb2.Progress(state=core_pb2.COMPLETED, status="", start=start, end=helper_functions.compute_timestamp())
            yield core_pb2.GetScoreSolutionResultsResponse(progress=msg, scores=send_scores)

    def FitSolution(self, request, context):
        """
        TA2-3 API call
        """
        logging.critical("Message received: FitSolution")
        request_id = str(uuid.uuid4())
        self._solution_score_map[request_id] = request
        return core_pb2.FitSolutionResponse(request_id = request_id)

    def GetFitSolutionResults(self, request, context):
        """
        TA2-3 API call
        """
        logging.critical("Message received: GetFitSolutionResults")
        request_id = request.request_id
        request_params = self._solution_score_map[request_id]
        start=helper_functions.compute_timestamp()

        solution_id = request_params.solution_id

        if solution_id not in self._solutions:
            logging.critical("GetFitSolutionResults: Solution %s not found!", solution_id)
            msg = core_pb2.Progress(state=core_pb2.ERRORED, status="", start=start, end=helper_functions.compute_timestamp())
            # Clean up
            self._solution_score_map.pop(request_id, None)
            yield core_pb2.GetFitSolutionResultsResponse(progress=msg, steps=[], exposed_outputs=[], fitted_solution_id=None)
        else:
            solution = self._solutions[solution_id]

            msg = core_pb2.Progress(state=core_pb2.RUNNING, status="", start=start, end=helper_functions.compute_timestamp())
            
            solution.create_pipeline_json(self._primitives, self._solutions) 

            inputs = self._get_inputs(solution.problem, request_params.inputs)
            try:
                output = solution.fit(inputs=inputs, solution_dict=self._solutions)
                logging.critical("Fit predictions with rows = %s", len(output))
            except:
                logging.critical("Exception in fit: %s", solution.primitives)
                logging.critical("Exception in fit: %s", sys.exc_info()[0])
                output = None

            result = None
            search_id_str = self._solution_to_search[solution_id]
            outputDir = os.environ['D3MOUTPUTDIR'] + "/" + search_id_str

            if output is not None:
                uri = util.write_predictions(output, outputDir + "/predictions", request_id)
                uri = 'file://{uri}'.format(uri=os.path.abspath(uri)) 
                result = value_pb2.Value(csv_uri=uri)
            else:
                result = value_pb2.Value(error = value_pb2.ValueError(message="Output is NULL"))

            yield core_pb2.GetFitSolutionResultsResponse(progress=msg, steps=[], exposed_outputs=[], fitted_solution_id=solution.id)

            msg = core_pb2.Progress(state=core_pb2.COMPLETED, status="", start=start, end=helper_functions.compute_timestamp())

            steps = []
            for i in range(solution.num_steps()):
                steps.append(core_pb2.StepProgress(progress=msg))

            exposed_outputs = {}
            last_step_output = None
            if request_params.expose_outputs is not None and len(request_params.expose_outputs) > 0:
                logging.critical("Fit %s", request_params.expose_outputs)
                for s in request_params.expose_outputs:
                    if 'feature_importances' in s:
                        (stepname, feature_imp) = solution.produce_feature_importances(None)
                        if feature_imp is not None:
                            uri = util.write_feature_importances(feature_imp, outputDir + "/predictions", request_id)
                            uri = 'file://{uri}'.format(uri=os.path.abspath(uri))
                            FI_result = value_pb2.Value(csv_uri=uri)
                            exposed_outputs[stepname] = FI_result
                        break
                for s in request_params.expose_outputs:
                    if 'outputs' in s:
                        last_step_output = s
                        break
                
            if last_step_output is None:
                last_step_output = solution.outputs[0][2]
            exposed_outputs[last_step_output] = result

            # Clean up
            self._solution_score_map.pop(request_id, None)

            yield core_pb2.GetFitSolutionResultsResponse(progress=msg, steps=steps, exposed_outputs=exposed_outputs, fitted_solution_id=solution.id)

    def ProduceSolution(self, request, context):
        """
        TA2-3 API call
        """
        logging.critical("Message received: ProduceSolution")
        request_id = str(uuid.uuid4())
        self._solution_score_map[request_id] = request

        return core_pb2.ProduceSolutionResponse(request_id = request_id)

    def GetProduceSolutionResults(self, request, context):
        """
        TA2-3 API call
        """
        logging.critical("Message received: GetProduceSolutionResults")
        request_id = request.request_id
        request_params = self._solution_score_map[request_id]
        start=helper_functions.compute_timestamp()

        solution_id = request_params.fitted_solution_id
        solution = self._solutions[solution_id]

        inputs = self._get_inputs(solution.problem, request_params.inputs)
        try:
            output = solution.produce(inputs=inputs, solution_dict=self._solutions)[0]
            logging.critical("Produce predictions with rows = %s", len(output))
        except:
            logging.critical("Exception in produce: %s", solution.primitives)
            logging.critical("Exception in produce: %s", sys.exc_info()[0])
            output = None
    
        result = None
        search_id_str = self._solution_to_search[solution_id]
        outputDir = os.environ['D3MOUTPUTDIR'] + "/" + search_id_str

        if output is not None:
            uri = util.write_predictions(output, outputDir + "/predictions", request_id)
            uri = 'file://{uri}'.format(uri=os.path.abspath(uri))
            result = value_pb2.Value(csv_uri=uri)
        else:
            result = value_pb2.Value(error = value_pb2.ValueError(message="Output is NULL"))

        self._solution_score_map.pop(request_id, None)

        msg = core_pb2.Progress(state=core_pb2.COMPLETED, status="", start=start, end=helper_functions.compute_timestamp())

        steps = []
        for i in range(solution.num_steps()):
            steps.append(core_pb2.StepProgress(progress=msg))

        exposed_outputs = {}
        last_step_output = None
        if request_params.expose_outputs is not None and len(request_params.expose_outputs) > 0:
            logging.critical("Produce %s", request_params.expose_outputs)
            for s in request_params.expose_outputs:
                if 'feature_importances' in s:
                    (stepname, feature_imp) = solution.produce_feature_importances(None)
                    if feature_imp is not None:
                        uri = util.write_feature_importances(feature_imp, outputDir + "/predictions", request_id)
                        uri = 'file://{uri}'.format(uri=os.path.abspath(uri))
                        FI_result = value_pb2.Value(csv_uri=uri)
                        exposed_outputs[stepname] = FI_result
                    break
            for s in request_params.expose_outputs:
                if 'outputs' in s:
                    last_step_output = s
                    break
        if last_step_output is None: 
            last_step_output = solution.outputs[0][2]
        exposed_outputs[last_step_output] = result

        yield core_pb2.GetProduceSolutionResultsResponse(progress=msg, steps=steps, exposed_outputs=exposed_outputs)

    def SolutionExport(self, request, context):
        """
        TA2-3 API call
        """
        logging.critical("Message received: SolutionExport")
        solution_id = request.solution_id
        rank = request.rank
        solution = self._solutions[solution_id]
        solution.rank = rank

        search_id_str = self._solution_to_search[solution_id]
        outputDir = os.environ['D3MOUTPUTDIR'] + "/" + search_id_str
        util.write_pipeline_json(solution, self._primitives, self._solutions, outputDir + "/pipelines_ranked", outputDir + "/subpipelines", rank=solution.rank)
        util.write_rank_file(solution, rank, outputDir + "/pipelines_ranked")

        return core_pb2.SolutionExportResponse()

    def UpdateProblem(self, request, context):
        """
        TA2-3 API call
        """
        logging.critical("Message received: UpdateProblem")

        return core_pb2.UpdateProblemResponse()

    def ListPrimitives(self, request, context):
        """
        TA2-3 API call
        """
        logging.critical("Message received: ListPrimitives")

        primitives = []
        for classname, p in self._primitives.items():
            primitives.append(primitive_pb2.Primitive(id=p.id, version=p.primitive_class.version, python_path=p.primitive_class.python_path, name=p.primitive_class.name, digest=None))
        return core_pb2.ListPrimitivesResponse(primitives=primitives)

    def Hello(self, request, context):
        """
        TA2-3 API call
        """
        logging.critical("Message received: Hello")
        version = core_pb2.DESCRIPTOR.GetOptions().Extensions[
                    core_pb2.protocol_version]
        return core_pb2.HelloResponse(user_agent="cmu_ta2",
        version=version,
        allowed_value_types = ['RAW', 'DATASET_URI', 'CSV_URI'],
        supported_extensions = [])
        
def add_to_server(server):
    core_pb2_grpc.add_CoreServicer_to_server(Core(), server)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: value.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='value.proto',
  package='',
  syntax='proto3',
  serialized_options=b'Z\010pipeline',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x0bvalue.proto\"\x1d\n\nValueError\x12\x0f\n\x07message\x18\x01 \x01(\t\"%\n\tValueList\x12\x18\n\x05items\x18\x01 \x03(\x0b\x32\t.ValueRaw\"j\n\tValueDict\x12$\n\x05items\x18\x01 \x03(\x0b\x32\x15.ValueDict.ItemsEntry\x1a\x37\n\nItemsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x18\n\x05value\x18\x02 \x01(\x0b\x32\t.ValueRaw:\x02\x38\x01\"\xbb\x01\n\x08ValueRaw\x12\x1a\n\x04null\x18\x01 \x01(\x0e\x32\n.NullValueH\x00\x12\x10\n\x06\x64ouble\x18\x02 \x01(\x01H\x00\x12\x0f\n\x05int64\x18\x03 \x01(\x03H\x00\x12\x0e\n\x04\x62ool\x18\x04 \x01(\x08H\x00\x12\x10\n\x06string\x18\x05 \x01(\tH\x00\x12\x0f\n\x05\x62ytes\x18\x06 \x01(\x0cH\x00\x12\x1a\n\x04list\x18\x07 \x01(\x0b\x32\n.ValueListH\x00\x12\x1a\n\x04\x64ict\x18\x08 \x01(\x0b\x32\n.ValueDictH\x00\x42\x05\n\x03raw\"\xb4\x01\n\x05Value\x12\x1c\n\x05\x65rror\x18\x01 \x01(\x0b\x32\x0b.ValueErrorH\x00\x12\x18\n\x03raw\x18\x02 \x01(\x0b\x32\t.ValueRawH\x00\x12\x15\n\x0b\x64\x61taset_uri\x18\x03 \x01(\tH\x00\x12\x11\n\x07\x63sv_uri\x18\x04 \x01(\tH\x00\x12\x14\n\npickle_uri\x18\x05 \x01(\tH\x00\x12\x15\n\x0bpickle_blob\x18\x06 \x01(\x0cH\x00\x12\x13\n\tplasma_id\x18\x07 \x01(\x0cH\x00\x42\x07\n\x05value*\x1b\n\tNullValue\x12\x0e\n\nNULL_VALUE\x10\x00\x42\nZ\x08pipelineb\x06proto3'
)

_NULLVALUE = _descriptor.EnumDescriptor(
  name='NullValue',
  full_name='NullValue',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='NULL_VALUE', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=566,
  serialized_end=593,
)
_sym_db.RegisterEnumDescriptor(_NULLVALUE)

NullValue = enum_type_wrapper.EnumTypeWrapper(_NULLVALUE)
NULL_VALUE = 0



_VALUEERROR = _descriptor.Descriptor(
  name='ValueError',
  full_name='ValueError',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='message', full_name='ValueError.message', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15,
  serialized_end=44,
)


_VALUELIST = _descriptor.Descriptor(
  name='ValueList',
  full_name='ValueList',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='items', full_name='ValueList.items', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=46,
  serialized_end=83,
)


_VALUEDICT_ITEMSENTRY = _descriptor.Descriptor(
  name='ItemsEntry',
  full_name='ValueDict.ItemsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='ValueDict.ItemsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='ValueDict.ItemsEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=136,
  serialized_end=191,
)

_VALUEDICT = _descriptor.Descriptor(
  name='ValueDict',
  full_name='ValueDict',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='items', full_name='ValueDict.items', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_VALUEDICT_ITEMSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=85,
  serialized_end=191,
)


_VALUERAW = _descriptor.Descriptor(
  name='ValueRaw',
  full_name='ValueRaw',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='null', full_name='ValueRaw.null', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='double', full_name='ValueRaw.double', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='int64', full_name='ValueRaw.int64', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='bool', full_name='ValueRaw.bool', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='string', full_name='ValueRaw.string', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='bytes', full_name='ValueRaw.bytes', index=5,
      number=6, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='list', full_name='ValueRaw.list', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dict', full_name='ValueRaw.dict', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='raw', full_name='ValueRaw.raw',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=194,
  serialized_end=381,
)


_VALUE = _descriptor.Descriptor(
  name='Value',
  full_name='Value',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='error', full_name='Value.error', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='raw', full_name='Value.raw', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dataset_uri', full_name='Value.dataset_uri', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='csv_uri', full_name='Value.csv_uri', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='pickle_uri', full_name='Value.pickle_uri', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='pickle_blob', full_name='Value.pickle_blob', index=5,
      number=6, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='plasma_id', full_name='Value.plasma_id', index=6,
      number=7, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='value', full_name='Value.value',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=384,
  serialized_end=564,
)

_VALUELIST.fields_by_name['items'].message_type = _VALUERAW
_VALUEDICT_ITEMSENTRY.fields_by_name['value'].message_type = _VALUERAW
_VALUEDICT_ITEMSENTRY.containing_type = _VALUEDICT
_VALUEDICT.fields_by_name['items'].message_type = _VALUEDICT_ITEMSENTRY
_VALUERAW.fields_by_name['null'].enum_type = _NULLVALUE
_VALUERAW.fields_by_name['list'].message_type = _VALUELIST
_VALUERAW.fields_by_name['dict'].message_type = _VALUEDICT
_VALUERAW.oneofs_by_name['raw'].fields.append(
  _VALUERAW.fields_by_name['null'])
_VALUERAW.fields_by_name['null'].containing_oneof = _VALUERAW.oneofs_by_name['raw']
_VALUERAW.oneofs_by_name['raw'].fields.append(
  _VALUERAW.fields_by_name['double'])
_VALUERAW.fields_by_name['double'].containing_oneof = _VALUERAW.oneofs_by_name['raw']
_VALUERAW.oneofs_by_name['raw'].fields.append(
  _VALUERAW.fields_by_name['int64'])
_VALUERAW.fields_by_name['int64'].containing_oneof = _VALUERAW.oneofs_by_name['raw']
_VALUERAW.oneofs_by_name['raw'].fields.append(
  _VALUERAW.fields_by_name['bool'])
_VALUERAW.fields_by_name['bool'].containing_oneof = _VALUERAW.oneofs_by_name['raw']
_VALUERAW.oneofs_by_name['raw'].fields.append(
  _VALUERAW.fields_by_name['string'])
_VALUERAW.fields_by_name['string'].containing_oneof = _VALUERAW.oneofs_by_name['raw']
_VALUERAW.oneofs_by_name['raw'].fields.append(
  _VALUERAW.fields_by_name['bytes'])
_VALUERAW.fields_by_name['bytes'].containing_oneof = _VALUERAW.oneofs_by_name['raw']
_VALUERAW.oneofs_by_name['raw'].fields.append(
  _VALUERAW.fields_by_name['list'])
_VALUERAW.fields_by_name['list'].containing_oneof = _VALUERAW.oneofs_by_name['raw']
_VALUERAW.oneofs_by_name['raw'].fields.append(
  _VALUERAW.fields_by_name['dict'])
_VALUERAW.fields_by_name['dict'].containing_oneof = _VALUERAW.oneofs_by_name['raw']
_VALUE.fields_by_name['error'].message_type = _VALUEERROR
_VALUE.fields_by_name['raw'].message_type = _VALUERAW
_VALUE.oneofs_by_name['value'].fields.append(
  _VALUE.fields_by_name['error'])
_VALUE.fields_by_name['error'].containing_oneof = _VALUE.oneofs_by_name['value']
_VALUE.oneofs_by_name['value'].fields.append(
  _VALUE.fields_by_name['raw'])
_VALUE.fields_by_name['raw'].containing_oneof = _VALUE.oneofs_by_name['value']
_VALUE.oneofs_by_name['value'].fields.append(
  _VALUE.fields_by_name['dataset_uri'])
_VALUE.fields_by_name['dataset_uri'].containing_oneof = _VALUE.oneofs_by_name['value']
_VALUE.oneofs_by_name['value'].fields.append(
  _VALUE.fields_by_name['csv_uri'])
_VALUE.fields_by_name['csv_uri'].containing_oneof = _VALUE.oneofs_by_name['value']
_VALUE.oneofs_by_name['value'].fields.append(
  _VALUE.fields_by_name['pickle_uri'])
_VALUE.fields_by_name['pickle_uri'].containing_oneof = _VALUE.oneofs_by_name['value']
_VALUE.oneofs_by_name['value'].fields.append(
  _VALUE.fields_by_name['pickle_blob'])
_VALUE.fields_by_name['pickle_blob'].containing_oneof = _VALUE.oneofs_by_name['value']
_VALUE.oneofs_by_name['value'].fields.append(
  _VALUE.fields_by_name['plasma_id'])
_VALUE.fields_by_name['plasma_id'].containing_oneof = _VALUE.oneofs_by_name['value']
DESCRIPTOR.message_types_by_name['ValueError'] = _VALUEERROR
DESCRIPTOR.message_types_by_name['ValueList'] = _VALUELIST
DESCRIPTOR.message_types_by_name['ValueDict'] = _VALUEDICT
DESCRIPTOR.message_types_by_name['ValueRaw'] = _VALUERAW
DESCRIPTOR.message_types_by_name['Value'] = _VALUE
DESCRIPTOR.enum_types_by_name['NullValue'] = _NULLVALUE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ValueError = _reflection.GeneratedProtocolMessageType('ValueError', (_message.Message,), {
  'DESCRIPTOR' : _VALUEERROR,
  '__module__' : 'value_pb2'
  # @@protoc_insertion_point(class_scope:ValueError)
  })
_sym_db.RegisterMessage(ValueError)

ValueList = _reflection.GeneratedProtocolMessageType('ValueList', (_message.Message,), {
  'DESCRIPTOR' : _VALUELIST,
  '__module__' : 'value_pb2'
  # @@protoc_insertion_point(class_scope:ValueList)
  })
_sym_db.RegisterMessage(ValueList)

ValueDict = _reflection.GeneratedProtocolMessageType('ValueDict', (_message.Message,), {

  'ItemsEntry' : _reflection.GeneratedProtocolMessageType('ItemsEntry', (_message.Message,), {
    'DESCRIPTOR' : _VALUEDICT_ITEMSENTRY,
    '__module__' : 'value_pb2'
    # @@protoc_insertion_point(class_scope:ValueDict.ItemsEntry)
    })
  ,
  'DESCRIPTOR' : _VALUEDICT,
  '__module__' : 'value_pb2'
  # @@protoc_insertion_point(class_scope:ValueDict)
  })
_sym_db.RegisterMessage(ValueDict)
_sym_db.RegisterMessage(ValueDict.ItemsEntry)

ValueRaw = _reflection.GeneratedProtocolMessageType('ValueRaw', (_message.Message,), {
  'DESCRIPTOR' : _VALUERAW,
  '__module__' : 'value_pb2'
  # @@protoc_insertion_point(class_scope:ValueRaw)
  })
_sym_db.RegisterMessage(ValueRaw)

Value = _reflection.GeneratedProtocolMessageType('Value', (_message.Message,), {
  'DESCRIPTOR' : _VALUE,
  '__module__' : 'value_pb2'
  # @@protoc_insertion_point(class_scope:Value)
  })
_sym_db.RegisterMessage(Value)


DESCRIPTOR._options = None
_VALUEDICT_ITEMSENTRY._options = None
# @@protoc_insertion_point(module_scope)




# File: search.py 
# Author(s): Saswati Ray
# Created: Wed Feb 17 06:44:20 EST 2021 
# Description:
# Acknowledgements:
# Copyright (c) 2021 Carnegie Mellon University
# This code is subject to the license terms contained in the code repo.

import os, sys, uuid
import logging
from multiprocessing import Pool, cpu_count
from timeit import default_timer as timer

# AutonML imports
import autonml.util as util
import autonml.auto_solutions as auto_solutions
import autonml.problem_pb2 as problem_pb2
import autonml.primitive_lib as primitive_lib

def rank_solutions(valid_solution_scores, problem_metric):
    """
    Return sorted list of multiple solutions.
    """
    # Sort solutions by their scores and rank them
    import operator
    sorted_x = sorted(valid_solution_scores.items(), key=operator.itemgetter(1))
    if util.invert_metric(problem_metric) is False:
        sorted_x.reverse()
    return sorted_x

def search_phase():
    """
    TA2 running in stand-alone search phase
    """
    overall_start = timer()
    inputDir = os.environ['D3MINPUTDIR']
    outputDir = os.environ['D3MOUTPUTDIR']
    timeout_env = os.environ['D3MTIMEOUT']
    num_cpus = (int)(os.environ['D3MCPU'])
    problemPath = os.environ['D3MPROBLEMPATH']

    logger = logging.getLogger()
    logger.setLevel(level=logging.ERROR)
    print("D3MINPUTDIR = ", inputDir)
    print("D3MOUTPUTDIR = ", outputDir)
    print("timeout = ", timeout_env)
    print("cpus = ", num_cpus)
    (dataset, task_name, problem_desc, metric, posLabel, keywords) = util.load_data_problem(inputDir, problemPath)
    testds = None #util.load_testdata(problemPath)

    async_message_thread = Pool(int(num_cpus))
    valid_solutions = {}
    valid_solution_scores = {}

    print("Metric = ", metric, " poslabel = ", posLabel)
    timeout_in_min = (int)(timeout_env)
    primitives = primitive_lib.load_primitives()
    print(task_name)

    problem_metric = "F1_MACRO"
    if metric == 'f1Macro':
        problem_metric = "F1_MACRO"
    elif metric == 'f1': 
        problem_metric = "F1"
    elif metric == 'accuracy':
        problem_metric = "ACCURACY"
    elif metric == 'meanSquaredError':
        problem_metric = "MEAN_SQUARED_ERROR"
    elif metric == 'rootMeanSquaredError':
        problem_metric = "ROOT_MEAN_SQUARED_ERROR"
    elif metric == 'meanAbsoluteError':
        problem_metric = "MEAN_ABSOLUTE_ERROR"
    elif 'rocAuc' in metric:
        problem_metric = "ROC_AUC"
    elif 'rSquared' in metric:
        problem_metric = "R_SQUARED"

    # Still run the normal pipeline even if augmentation
    start = timer()

    automl = auto_solutions.auto_solutions(task_name, None)
    solutions = automl.get_solutions(dataset, problem_metric, timeout_in_min)
    end = timer()
    time_used = end - start
    logging.critical("Time used up: %f seconds", time_used)

    inputs = []
    inputs.append(dataset)

    search_id_str = str(uuid.uuid4())
    outputDir = os.environ['D3MOUTPUTDIR'] + "/" + search_id_str
    util.initialize_for_search(outputDir)

    # Score potential solutions
    training_preds_dir = outputDir + "/training_predictions" 
    results = [async_message_thread.apply_async(evaluate_solution_score, (inputs, sol, primitives, problem_metric, posLabel, None, training_preds_dir,)) for sol in solutions]

    timeout = timeout_in_min * 60
    if timeout <= 0:
        timeout = None
    elif timeout > 60:
        timeout = timeout - 60

    if timeout is not None:
        timeout = timeout - time_used

    index = 0
    for r in results:
        try:
            start = timer()
            (score, optimal_params) = r.get(timeout=timeout)
            valid_solution_scores[index] = score
            if optimal_params is not None and len(optimal_params) > 0:
                solutions[index].set_hyperparams(None, optimal_params)
            end = timer()
            time_used = end - start
            timeout = timeout - time_used
            timeout = max(5, timeout)
        except:
            print(solutions[index].primitives)
            print(sys.exc_info()[0])
            print("Solution terminated: ", solutions[index].id)
        index = index + 1

    # Sort solutions by their scores and rank them
    sorted_x = rank_solutions(valid_solution_scores, problem_metric)

    rank = 1
    for (index, score) in sorted_x:
        id = solutions[index].id
        valid_solutions[id] = solutions[index]
        valid_solutions[id].rank = rank
        print("Rank ", rank)
        print("Score ", score)
        print(valid_solutions[id].primitives)
        rank = rank + 1

    num = 20
    if len(sorted_x) < 20:
        num = len(sorted_x)

    # Fit solutions and dump out files
    sorted_x = sorted_x[:num]
    testdata=[]
    testdata.append(testds)
    results = [async_message_thread.apply_async(fit_solution, (inputs, testdata, solutions[index], primitives, outputDir, problem_desc, score))
     for (index,score) in sorted_x]

    index = 0
    for r in results:
        try:
            valid=r.get(timeout=None)
        except:
            print(solutions[sorted_x[index][0]].primitives)
            print(sys.exc_info()[0])
            print("Fit Solution terminated: ", solutions[sorted_x[index][0]].id)
        index = index + 1

    overall_end = timer()
    logging.critical("Time used by AutonML : %f seconds", (overall_end - overall_start))

def evaluate_solution_score(inputs, solution, primitives, metric, posLabel, sol_dict, training_preds_dir=None):
    """
    Scores each potential solution
    Runs in a separate process
    """
    print("Evaluating ", solution.id)

    (score, optimal_params) = solution.score_solution(inputs=inputs, metric=metric, posLabel=posLabel,
                                primitive_dict=primitives, solution_dict=sol_dict, mode=training_preds_dir)

    return (score, optimal_params)

def fit_solution(inputs, testdata, solution, primitives, outputDir, problem_desc, score):
    """
    Fits each potential solution
    Runs in a separate process
    """
    print("Fitting ", solution.id)

    #output = solution.fit(inputs=inputs, solution_dict=None)
    #output = solution.produce(inputs=testdata, solution_dict=None)
    util.write_pipeline_json(solution, primitives, None, outputDir + "/pipelines_ranked", outputDir + "/subpipelines", rank=solution.rank, train_score=score)
    #util.write_pipeline_yaml(solution, outputDir + "/pipeline_runs", inputs, problem_desc)
    (stepname, feature_imp) = solution.produce_feature_importances(None)
    if feature_imp is not None:
        uri = util.write_feature_importances(feature_imp, outputDir + "/predictions", "12345")
        print(uri)
    return True




# File: solution_templates.py 
# Author(s): Saswati Ray
# Created: Wed Feb 17 06:44:20 EST 2021 
# Description:
# Acknowledgements:
# Copyright (c) 2021 Carnegie Mellon University
# This code is subject to the license terms contained in the code repo.

import os, sys

task_paths = {
'DISTILTEXT': ['d3m.primitives.data_transformation.denormalize.Common',
               'd3m.primitives.data_transformation.dataset_to_dataframe.Common',
               'd3m.primitives.schema_discovery.profiler.Common',
               'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',  # targets
               'd3m.primitives.data_transformation.column_parser.Common',
               'd3m.primitives.data_transformation.text_reader.Common',
               'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',
               'd3m.primitives.data_cleaning.imputer.SKlearn',
               'd3m.primitives.data_transformation.encoder.DistilTextEncoder'],

'TFIDFTEXT': ['d3m.primitives.data_transformation.denormalize.Common',
               'd3m.primitives.data_transformation.dataset_to_dataframe.Common',
               'd3m.primitives.schema_discovery.profiler.Common',
               'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',  # targets
               'd3m.primitives.data_transformation.column_parser.Common',
               'd3m.primitives.data_transformation.text_reader.Common',
               'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',
               'd3m.primitives.data_cleaning.imputer.SKlearn',
               'd3m.primitives.feature_extraction.tfidf_vectorizer.SKlearn'],

'COREXTEXT': ['d3m.primitives.data_transformation.denormalize.Common',
              'd3m.primitives.data_transformation.dataset_to_dataframe.Common',
              'd3m.primitives.schema_discovery.profiler.Common',
              'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',  # targets
              'd3m.primitives.data_transformation.column_parser.Common',
              'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',
              'd3m.primitives.data_cleaning.imputer.SKlearn',
              'd3m.primitives.feature_construction.corex_text.DSBOX'],

'TIMESERIES': ['d3m.primitives.data_transformation.denormalize.Common',
               'd3m.primitives.data_transformation.dataset_to_dataframe.Common',
               'd3m.primitives.schema_discovery.profiler.Common',
               'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',  # targets
               'd3m.primitives.data_transformation.time_series_to_list.DSBOX',
               'd3m.primitives.feature_extraction.random_projection_timeseries_featurization.DSBOX',
               'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common'],

'TIMESERIES2': ['d3m.primitives.data_transformation.time_series_formatter.DistilTimeSeriesFormatter',
                'd3m.primitives.data_transformation.dataset_to_dataframe.Common',
                'd3m.primitives.data_transformation.dataset_to_dataframe.Common',
                'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common'],

'TIMESERIES3': ['d3m.primitives.data_transformation.denormalize.Common',
                'd3m.primitives.time_series_classification.convolutional_neural_net.LSTM_FCN'],

'ISI_PIPELINE': ['d3m.primitives.data_transformation.denormalize.Common',
                 'd3m.primitives.data_transformation.dataset_to_dataframe.Common',
                 'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',  # targets
                 'd3m.primitives.feature_construction.gcn_mixhop.DSBOX'],

'IMAGE2': ['d3m.primitives.data_transformation.denormalize.Common',
           'd3m.primitives.data_transformation.dataset_to_dataframe.Common',
           'd3m.primitives.schema_discovery.profiler.Common',
           'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',  # targets
           'd3m.primitives.data_transformation.dataframe_to_tensor.DSBOX',
           'd3m.primitives.feature_extraction.resnet50_image_feature.DSBOX'],

'IMAGE': ['d3m.primitives.data_transformation.denormalize.Common',
          'd3m.primitives.data_transformation.dataset_to_dataframe.Common',
          'd3m.primitives.schema_discovery.profiler.Common',
          'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',  # targets
          'd3m.primitives.data_transformation.image_reader.Common',
          'd3m.primitives.data_transformation.column_parser.Common',
          'd3m.primitives.feature_extraction.image_transfer.DistilImageTransfer'],

'VIDEO': ['d3m.primitives.data_transformation.denormalize.Common',
          'd3m.primitives.data_transformation.dataset_to_dataframe.Common',
          'd3m.primitives.schema_discovery.profiler.Common',
          'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',
          'd3m.primitives.data_transformation.column_parser.Common',
          'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',
          'd3m.primitives.feature_extraction.resnext101_kinetics_video_features.VideoFeaturizer'],

'CLASSIFICATION': ['d3m.primitives.data_transformation.denormalize.Common',
                   'd3m.primitives.data_transformation.dataset_to_dataframe.Common',
                   'd3m.primitives.schema_discovery.profiler.Common',
                   'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',  # targets
                   'd3m.primitives.data_transformation.column_parser.Common',
                   'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',
                   'd3m.primitives.data_cleaning.imputer.SKlearn'],

'CONDITIONER': ['d3m.primitives.data_transformation.denormalize.Common',
                'd3m.primitives.data_transformation.dataset_to_dataframe.Common',
                'd3m.primitives.schema_discovery.profiler.Common',
                'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',  # targets
                'd3m.primitives.data_transformation.column_parser.Common',
                'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',
                'd3m.primitives.data_transformation.conditioner.Conditioner',
                'd3m.primitives.feature_extraction.feature_agglomeration.SKlearn'],

'SEMISUPERVISED': ['d3m.primitives.data_transformation.denormalize.Common',
                   'd3m.primitives.data_transformation.dataset_to_dataframe.Common',
                   'd3m.primitives.schema_discovery.profiler.Common',
                   'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',  # targets
                   'd3m.primitives.data_transformation.column_parser.Common',
                   'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',
                   'd3m.primitives.data_cleaning.imputer.SKlearn'],

'SEMISUPERVISED_HDB': ['d3m.primitives.data_transformation.denormalize.Common',
                       'd3m.primitives.data_transformation.dataset_to_dataframe.Common',
                       'd3m.primitives.schema_discovery.profiler.Common',
                       'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',  # targets
                       'd3m.primitives.data_transformation.column_parser.Common',
                       'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',
                       'd3m.primitives.clustering.hdbscan.Hdbscan',
                       'd3m.primitives.data_cleaning.imputer.SKlearn'],

'ARIMA_FORECASTING': ['d3m.primitives.data_transformation.denormalize.Common',
                      'd3m.primitives.data_transformation.dataset_to_dataframe.Common',
                      'd3m.primitives.schema_discovery.profiler.Common',
                      'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',  # targets
                      'd3m.primitives.data_transformation.column_parser.Common',
                      'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',
                      'd3m.primitives.time_series_forecasting.arima.DSBOX'],

'VAR_FORECASTING': ['d3m.primitives.data_transformation.denormalize.Common',
                    'd3m.primitives.data_transformation.dataset_to_dataframe.Common',
                    'd3m.primitives.schema_discovery.profiler.Common',
                    'd3m.primitives.data_transformation.column_parser.Common',
                    'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',  # targets
                    'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',
                    'd3m.primitives.time_series_forecasting.vector_autoregression.VAR'],

'LSTM_FORECASTING': ['d3m.primitives.data_transformation.denormalize.Common',
                      'd3m.primitives.data_transformation.dataset_to_dataframe.Common',
                      'd3m.primitives.schema_discovery.profiler.Common',
                      'd3m.primitives.data_transformation.column_parser.Common',
                      'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',  # targets
                      'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',
                      'd3m.primitives.time_series_forecasting.lstm.DeepAR'],

'ESRNN_FORECASTING': ['d3m.primitives.data_transformation.denormalize.Common',
                      'd3m.primitives.data_transformation.dataset_to_dataframe.Common',
                      'd3m.primitives.schema_discovery.profiler.Common',
                      'd3m.primitives.data_transformation.column_parser.Common',
                      'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',  # targets
                      'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',
                      'd3m.primitives.data_cleaning.imputer.SKlearn',
                      'd3m.primitives.data_transformation.grouping_field_compose.Common',
                      'd3m.primitives.time_series_forecasting.esrnn.RNN',
                      'd3m.primitives.data_transformation.construct_predictions.Common'],

'NBEATS_FORECASTING': ['d3m.primitives.data_transformation.denormalize.Common',
                       'd3m.primitives.data_transformation.dataset_to_dataframe.Common',
                       'd3m.primitives.schema_discovery.profiler.Common',
                       'd3m.primitives.data_transformation.column_parser.Common',
                       'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',  # targets
                       'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',
                       'd3m.primitives.data_cleaning.imputer.SKlearn',
                       'd3m.primitives.data_transformation.grouping_field_compose.Common',
                       'd3m.primitives.time_series_forecasting.nbeats.DeepNeuralNetwork',
                       'd3m.primitives.data_transformation.construct_predictions.Common'],

'DISTIL_NBEATS': ['d3m.primitives.data_transformation.denormalize.Common',
                  'd3m.primitives.data_transformation.dataset_to_dataframe.Common',
                  'd3m.primitives.schema_discovery.profiler.Common',
                  'd3m.primitives.data_transformation.column_parser.Common',
                  'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',  # targets
                  'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',
                  'd3m.primitives.time_series_forecasting.feed_forward_neural_net.NBEATS',
                  'd3m.primitives.data_transformation.construct_predictions.Common'],

'REGRESSION': ['d3m.primitives.data_transformation.denormalize.Common',
               'd3m.primitives.data_transformation.dataset_to_dataframe.Common',
               'd3m.primitives.schema_discovery.profiler.Common',
               'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',  # targets
               'd3m.primitives.data_transformation.column_parser.Common',
               'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',
               'd3m.primitives.data_cleaning.imputer.SKlearn'],

'FORE_REGRESSION': ['d3m.primitives.data_transformation.denormalize.Common',
                    'd3m.primitives.data_transformation.dataset_to_dataframe.Common',
                    'd3m.primitives.schema_discovery.profiler.Common',
                    'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',  # targets
                    'd3m.primitives.data_transformation.column_parser.Common',
                    'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',
                    'd3m.primitives.data_cleaning.imputer.SKlearn'],

'PIPELINE_RPI': ['d3m.primitives.data_transformation.denormalize.Common',
                 'd3m.primitives.data_transformation.dataset_to_dataframe.Common',
                 'd3m.primitives.schema_discovery.profiler.Common',
                 'd3m.primitives.data_transformation.column_parser.Common',
                 'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',  # targets
                 'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common'],

'NOTUNE_PIPELINE_RPI': ['d3m.primitives.data_transformation.denormalize.Common',
                        'd3m.primitives.data_transformation.dataset_to_dataframe.Common',
                        'd3m.primitives.schema_discovery.profiler.Common',
                        'd3m.primitives.data_transformation.column_parser.Common',
                        'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',  # targets
                        'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',
                        'd3m.primitives.feature_selection.joint_mutual_information.AutoRPI',
                        'd3m.primitives.data_cleaning.imputer.SKlearn'],

'CLUSTERING': ['d3m.primitives.data_transformation.dataset_to_dataframe.Common',
               'd3m.primitives.schema_discovery.profiler.Common',
               'd3m.primitives.data_transformation.column_parser.Common',
               'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',
               'd3m.primitives.clustering.k_means.Fastlvm',
               'd3m.primitives.data_transformation.construct_predictions.Common'],

'GRAPHMATCHING': ['d3m.primitives.data_transformation.load_graphs.DistilGraphLoader',
                  'd3m.primitives.graph_matching.seeded_graph_matching.DistilSeededGraphMatcher'],

'GRAPHMATCHING2': ['d3m.primitives.graph_matching.seeded_graph_matching.JHU'],

'GRAPHMATCHING3': ['d3m.primitives.data_transformation.dataset_to_dataframe.Common',
                   'd3m.primitives.data_transformation.dataset_to_dataframe.Common',
                   'd3m.primitives.data_transformation.dataset_to_dataframe.Common',
                   'd3m.primitives.graph_matching.euclidean_nomination.JHU'],

'COLLABORATIVEFILTERING': ['d3m.primitives.data_transformation.denormalize.Common',
                           'd3m.primitives.data_transformation.dataset_to_dataframe.Common',
                           'd3m.primitives.schema_discovery.profiler.Common',
                           'd3m.primitives.data_transformation.column_parser.Common',
                           'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',
                           'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',
                           'd3m.primitives.collaborative_filtering.link_prediction.DistilCollaborativeFiltering',
                           'd3m.primitives.data_transformation.construct_predictions.Common'],

'VERTEXCLASSIFICATION3': ['d3m.primitives.data_transformation.load_edgelist.DistilEdgeListLoader',
                         'd3m.primitives.vertex_nomination.seeded_graph_matching.DistilVertexNomination'],

'VERTEXCLASSIFICATION2': ['d3m.primitives.data_transformation.load_graphs.JHU',
                          'd3m.primitives.data_preprocessing.largest_connected_component.JHU',
                          'd3m.primitives.data_transformation.adjacency_spectral_embedding.JHU',
                          'd3m.primitives.classification.gaussian_classification.JHU'],

'VERTEXCLASSIFICATION': ['d3m.primitives.data_transformation.load_single_graph.DistilSingleGraphLoader',
                         'd3m.primitives.vertex_nomination.seeded_graph_matching.DistilVertexNomination'],

'OBJECTDETECTION': ['d3m.primitives.data_transformation.denormalize.Common',
                    'd3m.primitives.data_transformation.dataset_to_dataframe.Common',
                    'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',
                    'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',
                    'd3m.primitives.feature_extraction.yolo.DSBOX'],

'OBJECTDETECTION2': ['d3m.primitives.data_transformation.denormalize.Common',
                     'd3m.primitives.data_transformation.dataset_to_dataframe.Common',
                     'd3m.primitives.object_detection.retina_net.ObjectDetectionRN'],

'LINKPREDICTION2': ['d3m.primitives.link_prediction.data_conversion.JHU', 
                    'd3m.primitives.data_transformation.adjacency_spectral_embedding.JHU',
                    'd3m.primitives.link_prediction.rank_classification.JHU'],

'LINKPREDICTION': ['d3m.primitives.data_transformation.load_single_graph.DistilSingleGraphLoader',
                    'd3m.primitives.link_prediction.link_prediction.DistilLinkPrediction'],

'COMMUNITYDETECTION': ['d3m.primitives.data_transformation.load_single_graph.DistilSingleGraphLoader',
                       'd3m.primitives.community_detection.parser.DistilCommunityDetection'],

'COMMUNITYDETECTION2': ['d3m.primitives.data_transformation.load_graphs.JHU',
	                'd3m.primitives.data_preprocessing.largest_connected_component.JHU',
                        'd3m.primitives.data_transformation.adjacency_spectral_embedding.JHU',
	                'd3m.primitives.graph_clustering.gaussian_clustering.JHU'],

'IMVADIO': ['d3m.primitives.data_transformation.denormalize.Common',
          'd3m.primitives.data_transformation.dataset_to_dataframe.Common',
          'd3m.primitives.schema_discovery.profiler.Common',
          'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',  # targets
          'd3m.primitives.data_transformation.column_parser.Common',
          'd3m.primitives.data_transformation.text_reader.Common',
          'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',
          'd3m.primitives.data_transformation.encoder.DistilTextEncoder'],

'REMOTESENSING': ['d3m.primitives.data_transformation.denormalize.Common',
                  'd3m.primitives.data_transformation.dataset_to_dataframe.Common',
                  'd3m.primitives.data_transformation.satellite_image_loader.DistilSatelliteImageLoader',
                  'd3m.primitives.data_transformation.column_parser.Common',
                  'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common', # images
                  'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',  # targets
                  'd3m.primitives.remote_sensing.remote_sensing_pretrained.RemoteSensingPretrained'],

'AUDIO': ['d3m.primitives.data_transformation.audio_reader.DistilAudioDatasetLoader',
          'd3m.primitives.data_transformation.column_parser.Common',
          'd3m.primitives.data_transformation.extract_columns_by_semantic_types.Common',  # targets
          'd3m.primitives.feature_extraction.audio_transfer.DistilAudioTransfer']}

classifiers = [#'d3m.primitives.classification.multinomial_naive_bayes.SKlearn',
               'd3m.primitives.classification.linear_discriminant_analysis.SKlearn',
               'd3m.primitives.classification.quadratic_discriminant_analysis.SKlearn',
               'd3m.primitives.classification.logistic_regression.SKlearn',
               'd3m.primitives.classification.ada_boost.SKlearn',
               'd3m.primitives.classification.linear_svc.SKlearn',
               'd3m.primitives.classification.extra_trees.SKlearn',
               'd3m.primitives.classification.random_forest.SKlearn',
               'd3m.primitives.classification.bagging.SKlearn',
               'd3m.primitives.classification.svc.SKlearn',
               'd3m.primitives.classification.passive_aggressive.SKlearn',
               'd3m.primitives.classification.mlp.SKlearn',
               'd3m.primitives.classification.xgboost_gbtree.Common',
               'd3m.primitives.classification.gradient_boosting.SKlearn']

regressors = ['d3m.primitives.regression.ridge.SKlearn',
              'd3m.primitives.regression.lasso.SKlearn',
              'd3m.primitives.regression.elastic_net.SKlearn',
              'd3m.primitives.regression.lasso_cv.SKlearn',
              'd3m.primitives.regression.ada_boost.SKlearn',
              'd3m.primitives.regression.linear_svr.SKlearn',
              'd3m.primitives.regression.random_forest.SKlearn',
              'd3m.primitives.regression.extra_trees.SKlearn',
              'd3m.primitives.regression.xgboost_gbtree.Common',
              'd3m.primitives.regression.bagging.SKlearn',
              'd3m.primitives.regression.mlp.SKlearn',
              'd3m.primitives.regression.gradient_boosting.SKlearn']

regressors_rpi = ['d3m.primitives.regression.random_forest.SKlearn',
                  'd3m.primitives.regression.extra_trees.SKlearn',
                  'd3m.primitives.regression.gradient_boosting.SKlearn']

classifiers_rpi = ['d3m.primitives.classification.random_forest.SKlearn',
                   'd3m.primitives.classification.extra_trees.SKlearn',
                   'd3m.primitives.classification.gradient_boosting.SKlearn',
                   'd3m.primitives.classification.linear_discriminant_analysis.SKlearn']

sslVariants = ['d3m.primitives.classification.gradient_boosting.SKlearn',
               'd3m.primitives.classification.extra_trees.SKlearn',
               'd3m.primitives.classification.random_forest.SKlearn',
               'd3m.primitives.classification.bagging.SKlearn',
               'd3m.primitives.classification.linear_svc.SKlearn',
               'd3m.primitives.classification.svc.SKlearn']




# File: setup.py 
# Author(s): Vedant Sanil
# Created: Wed Feb 17 11:49:20 EST 2022 
# Description:
# Acknowledgements:
# Copyright (c) 2022 Carnegie Mellon University
# This code is subject to the license terms contained in the code repo.

import io
import os
import sys
import shutil
import setuptools
import subprocess
from glob import glob
from distutils.command.build_py import build_py

from autonml import __VERSION__

NAME = 'autonml'
VERSION = __VERSION__

with io.open('README.md', 'r', encoding="utf-8") as fp:
    description = fp.read()

with open('requirements.txt', 'r') as reqfile:
    req = [line.strip() for line in reqfile if line and not line.startswith('#')]

def run(args):
    subprocess.run(args, stdout=sys.stdout, stderr=sys.stdout, check=True, encoding='utf8')
    sys.stdout.flush()

pkgs = [elem.replace('autonml/', '') for elem in glob('autonml/static/*', recursive=True) if os.path.isfile(elem)]

setuptools.setup(
    name=NAME,
    version=VERSION,
    install_requires=req,
    description=r"AutonML : CMU's AutoML System",
    long_description=description,
    package_data={NAME:pkgs},
    packages=setuptools.find_packages(),
    entry_points={
        'console_scripts': [
            'autonml_main=autonml.main:main_run',
            'automl_search=autonml.main:main_search',
            'create_d3m_dataset=autonml.create_d3m_dataset:main']},
    python_requires=">=3.6",
    include_package_data=True,
    dependency_links = [ 
        'https://github.com/uncharted-distil/distil-primitives.git@666233600447f7c4cc09fbd0e59c87ca5b842011#egg=distil-primitives',
        'https://gitlab.com/datadrivendiscovery/contrib/kungfuai-primitives.git@49acb225bb6994d3dfaffdf3b7761395423680a4#egg=kf-d3m-primitives'
    ],
    author='Saswati Ray, Andrew Williams, Vedant Sanil',
    maintainer='Andrew Williams, Vedant Sanil',
    maintainer_email='awillia2@andrew.cmu.edu, vsanil@andrew.cmu.edu',
    keywords=['datadrivendiscovery', 'automl', 'd3m', 'ta2', 'cmu'],
    license='Apache-2.0',
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: Apache Software License',
        'Topic :: Scientific/Engineering']
)


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: problem.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='problem.proto',
  package='',
  syntax='proto3',
  serialized_options=b'Z\010pipeline',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\rproblem.proto\"H\n\x18ProblemPerformanceMetric\x12\x0e\n\x06metric\x18\x01 \x01(\t\x12\t\n\x01k\x18\x02 \x01(\x05\x12\x11\n\tpos_label\x18\x03 \x01(\t\"X\n\x07Problem\x12\x15\n\rtask_keywords\x18\x08 \x03(\t\x12\x36\n\x13performance_metrics\x18\x07 \x03(\x0b\x32\x19.ProblemPerformanceMetric\"~\n\rProblemTarget\x12\x14\n\x0ctarget_index\x18\x01 \x01(\x05\x12\x13\n\x0bresource_id\x18\x02 \x01(\t\x12\x14\n\x0c\x63olumn_index\x18\x03 \x01(\x05\x12\x13\n\x0b\x63olumn_name\x18\x04 \x01(\t\x12\x17\n\x0f\x63lusters_number\x18\x05 \x01(\x05\"v\n\x15ProblemPrivilegedData\x12\x1d\n\x15privileged_data_index\x18\x01 \x01(\x05\x12\x13\n\x0bresource_id\x18\x02 \x01(\t\x12\x14\n\x0c\x63olumn_index\x18\x03 \x01(\x05\x12\x13\n\x0b\x63olumn_name\x18\x04 \x01(\t\"k\n\x12\x46orecastingHorizon\x12\x13\n\x0bresource_id\x18\x01 \x01(\t\x12\x14\n\x0c\x63olumn_index\x18\x02 \x01(\x05\x12\x13\n\x0b\x63olumn_name\x18\x03 \x01(\t\x12\x15\n\rhorizon_value\x18\x04 \x01(\x01\"\xa6\x01\n\x0cProblemInput\x12\x12\n\ndataset_id\x18\x01 \x01(\t\x12\x1f\n\x07targets\x18\x02 \x03(\x0b\x32\x0e.ProblemTarget\x12/\n\x0fprivileged_data\x18\x03 \x03(\x0b\x32\x16.ProblemPrivilegedData\x12\x30\n\x13\x66orecasting_horizon\x18\x04 \x01(\x0b\x32\x13.ForecastingHorizon\"4\n\x10\x44\x61taAugmentation\x12\x0e\n\x06\x64omain\x18\x01 \x03(\t\x12\x10\n\x08keywords\x18\x02 \x03(\t\"\xe1\x01\n\x12ProblemDescription\x12\x19\n\x07problem\x18\x01 \x01(\x0b\x32\x08.Problem\x12\x1d\n\x06inputs\x18\x02 \x03(\x0b\x32\r.ProblemInput\x12\n\n\x02id\x18\x03 \x01(\t\x12\x0f\n\x07version\x18\x04 \x01(\t\x12\x0c\n\x04name\x18\x05 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x06 \x01(\t\x12\x0e\n\x06\x64igest\x18\x07 \x01(\t\x12,\n\x11\x64\x61ta_augmentation\x18\x08 \x03(\x0b\x32\x11.DataAugmentation\x12\x13\n\x0bother_names\x18\t \x03(\tB\nZ\x08pipelineb\x06proto3'
)




_PROBLEMPERFORMANCEMETRIC = _descriptor.Descriptor(
  name='ProblemPerformanceMetric',
  full_name='ProblemPerformanceMetric',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='metric', full_name='ProblemPerformanceMetric.metric', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='k', full_name='ProblemPerformanceMetric.k', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='pos_label', full_name='ProblemPerformanceMetric.pos_label', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17,
  serialized_end=89,
)


_PROBLEM = _descriptor.Descriptor(
  name='Problem',
  full_name='Problem',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='task_keywords', full_name='Problem.task_keywords', index=0,
      number=8, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='performance_metrics', full_name='Problem.performance_metrics', index=1,
      number=7, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=91,
  serialized_end=179,
)


_PROBLEMTARGET = _descriptor.Descriptor(
  name='ProblemTarget',
  full_name='ProblemTarget',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='target_index', full_name='ProblemTarget.target_index', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='resource_id', full_name='ProblemTarget.resource_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='column_index', full_name='ProblemTarget.column_index', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='column_name', full_name='ProblemTarget.column_name', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='clusters_number', full_name='ProblemTarget.clusters_number', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=181,
  serialized_end=307,
)


_PROBLEMPRIVILEGEDDATA = _descriptor.Descriptor(
  name='ProblemPrivilegedData',
  full_name='ProblemPrivilegedData',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='privileged_data_index', full_name='ProblemPrivilegedData.privileged_data_index', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='resource_id', full_name='ProblemPrivilegedData.resource_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='column_index', full_name='ProblemPrivilegedData.column_index', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='column_name', full_name='ProblemPrivilegedData.column_name', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=309,
  serialized_end=427,
)


_FORECASTINGHORIZON = _descriptor.Descriptor(
  name='ForecastingHorizon',
  full_name='ForecastingHorizon',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='resource_id', full_name='ForecastingHorizon.resource_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='column_index', full_name='ForecastingHorizon.column_index', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='column_name', full_name='ForecastingHorizon.column_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='horizon_value', full_name='ForecastingHorizon.horizon_value', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=429,
  serialized_end=536,
)


_PROBLEMINPUT = _descriptor.Descriptor(
  name='ProblemInput',
  full_name='ProblemInput',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='dataset_id', full_name='ProblemInput.dataset_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='targets', full_name='ProblemInput.targets', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='privileged_data', full_name='ProblemInput.privileged_data', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='forecasting_horizon', full_name='ProblemInput.forecasting_horizon', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=539,
  serialized_end=705,
)


_DATAAUGMENTATION = _descriptor.Descriptor(
  name='DataAugmentation',
  full_name='DataAugmentation',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='domain', full_name='DataAugmentation.domain', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='keywords', full_name='DataAugmentation.keywords', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=707,
  serialized_end=759,
)


_PROBLEMDESCRIPTION = _descriptor.Descriptor(
  name='ProblemDescription',
  full_name='ProblemDescription',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='problem', full_name='ProblemDescription.problem', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='inputs', full_name='ProblemDescription.inputs', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='id', full_name='ProblemDescription.id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='version', full_name='ProblemDescription.version', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name', full_name='ProblemDescription.name', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='ProblemDescription.description', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='digest', full_name='ProblemDescription.digest', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data_augmentation', full_name='ProblemDescription.data_augmentation', index=7,
      number=8, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='other_names', full_name='ProblemDescription.other_names', index=8,
      number=9, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=762,
  serialized_end=987,
)

_PROBLEM.fields_by_name['performance_metrics'].message_type = _PROBLEMPERFORMANCEMETRIC
_PROBLEMINPUT.fields_by_name['targets'].message_type = _PROBLEMTARGET
_PROBLEMINPUT.fields_by_name['privileged_data'].message_type = _PROBLEMPRIVILEGEDDATA
_PROBLEMINPUT.fields_by_name['forecasting_horizon'].message_type = _FORECASTINGHORIZON
_PROBLEMDESCRIPTION.fields_by_name['problem'].message_type = _PROBLEM
_PROBLEMDESCRIPTION.fields_by_name['inputs'].message_type = _PROBLEMINPUT
_PROBLEMDESCRIPTION.fields_by_name['data_augmentation'].message_type = _DATAAUGMENTATION
DESCRIPTOR.message_types_by_name['ProblemPerformanceMetric'] = _PROBLEMPERFORMANCEMETRIC
DESCRIPTOR.message_types_by_name['Problem'] = _PROBLEM
DESCRIPTOR.message_types_by_name['ProblemTarget'] = _PROBLEMTARGET
DESCRIPTOR.message_types_by_name['ProblemPrivilegedData'] = _PROBLEMPRIVILEGEDDATA
DESCRIPTOR.message_types_by_name['ForecastingHorizon'] = _FORECASTINGHORIZON
DESCRIPTOR.message_types_by_name['ProblemInput'] = _PROBLEMINPUT
DESCRIPTOR.message_types_by_name['DataAugmentation'] = _DATAAUGMENTATION
DESCRIPTOR.message_types_by_name['ProblemDescription'] = _PROBLEMDESCRIPTION
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ProblemPerformanceMetric = _reflection.GeneratedProtocolMessageType('ProblemPerformanceMetric', (_message.Message,), {
  'DESCRIPTOR' : _PROBLEMPERFORMANCEMETRIC,
  '__module__' : 'problem_pb2'
  # @@protoc_insertion_point(class_scope:ProblemPerformanceMetric)
  })
_sym_db.RegisterMessage(ProblemPerformanceMetric)

Problem = _reflection.GeneratedProtocolMessageType('Problem', (_message.Message,), {
  'DESCRIPTOR' : _PROBLEM,
  '__module__' : 'problem_pb2'
  # @@protoc_insertion_point(class_scope:Problem)
  })
_sym_db.RegisterMessage(Problem)

ProblemTarget = _reflection.GeneratedProtocolMessageType('ProblemTarget', (_message.Message,), {
  'DESCRIPTOR' : _PROBLEMTARGET,
  '__module__' : 'problem_pb2'
  # @@protoc_insertion_point(class_scope:ProblemTarget)
  })
_sym_db.RegisterMessage(ProblemTarget)

ProblemPrivilegedData = _reflection.GeneratedProtocolMessageType('ProblemPrivilegedData', (_message.Message,), {
  'DESCRIPTOR' : _PROBLEMPRIVILEGEDDATA,
  '__module__' : 'problem_pb2'
  # @@protoc_insertion_point(class_scope:ProblemPrivilegedData)
  })
_sym_db.RegisterMessage(ProblemPrivilegedData)

ForecastingHorizon = _reflection.GeneratedProtocolMessageType('ForecastingHorizon', (_message.Message,), {
  'DESCRIPTOR' : _FORECASTINGHORIZON,
  '__module__' : 'problem_pb2'
  # @@protoc_insertion_point(class_scope:ForecastingHorizon)
  })
_sym_db.RegisterMessage(ForecastingHorizon)

ProblemInput = _reflection.GeneratedProtocolMessageType('ProblemInput', (_message.Message,), {
  'DESCRIPTOR' : _PROBLEMINPUT,
  '__module__' : 'problem_pb2'
  # @@protoc_insertion_point(class_scope:ProblemInput)
  })
_sym_db.RegisterMessage(ProblemInput)

DataAugmentation = _reflection.GeneratedProtocolMessageType('DataAugmentation', (_message.Message,), {
  'DESCRIPTOR' : _DATAAUGMENTATION,
  '__module__' : 'problem_pb2'
  # @@protoc_insertion_point(class_scope:DataAugmentation)
  })
_sym_db.RegisterMessage(DataAugmentation)

ProblemDescription = _reflection.GeneratedProtocolMessageType('ProblemDescription', (_message.Message,), {
  'DESCRIPTOR' : _PROBLEMDESCRIPTION,
  '__module__' : 'problem_pb2'
  # @@protoc_insertion_point(class_scope:ProblemDescription)
  })
_sym_db.RegisterMessage(ProblemDescription)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)




# File: auto_solutions.py 
# Author(s): Saswati Ray
# Created: Wed Feb 17 06:44:20 EST 2021 
# Description:
# Acknowledgements:
# Copyright (c) 2021 Carnegie Mellon University
# This code is subject to the license terms contained in the code repo.

import os, copy, uuid, sys
import logging
import numpy as np
import traceback

# AutonML imports
import autonml.util as util
import autonml.solutiondescription as solutiondescription
import autonml.solution_templates as solution_templates
import autonml.default_hyperparams as default_hyperparams
import autonml.helper_functions as helper_functions

def is_primitive_reasonable(python_path, rows, total_cols, types_present, task_name, metric):
    """
    Rule-based elimination based on ML task, dataset size (rows, cols).
    Is primitive reasonable to evaluate for the ML task, dataset size?
    Eliminate expensive solutions here!!!
    """
    if 'ROC_AUC' in metric and \
       ('passive' in python_path or \
        'discriminant' in python_path or \
        'svc' in python_path):
        return False

    if (total_cols > 1000 or rows > 100000) and 'xgboost' in python_path:
        return False

    if total_cols > 2000 and ('discriminant' in python_path or 'passive' in python_path):
        return False

    if (total_cols > 100 or rows > 100000) and 'k_neighbors' in python_path:
        return False

    if complex_types_present(types_present) is True and \
       ('ada_boost' in python_path or \
        'passive_aggressive' in python_path or \
        'xgboost' in python_path or \
        'gradient_boosting' in python_path or \
        'quadratic' in python_path or \
        'mlp' in python_path or \
        'bagging' in python_path):
        return False

    # Forecasting
    if task_name == "FORE_REGRESSION" and 'mlp' in python_path:
        return False

    # SVM gets extremely expensive (memory) for >10k samples!!!
    if rows > 10000 and 'classification.svc.SKlearn' in python_path:
        return False

    if rows >= 100000 and ('gradient_boosting' in python_path or 'mlp.SKlearn' in python_path or 'ada_boost' in python_path or 'passive_aggressive' in python_path):
        return False

    if rows > 1000000 and ('bagging' in python_path or 'elastic' in python_path or 'lasso_cv' in python_path):
        return False

    return True

def complex_types_present(types_present):
    """
    Does datatypes contain multi-media??
    """
    if types_present is None:
        return False
    if 'AUDIO' in types_present or \
       'VIDEO' in types_present or \
       'IMAGE' in types_present:
        return True
    return False

class auto_solutions(object):
    """
    Main class representing the AutoML (pipeline constructor) for TA2.
    A "pipeline" and "solution" will mean the SAME thing in D3M.
    Creates a suite of suitable pipelines for a given dataset/problem type.
    Pipelines/Solutions are only created using templates and not run here.
    In case of tasks with multiple pipelines (classification/regression/SSL), primary data processing/featurization steps are
    run only once and cached for all the pipelines.
    """
    def __init__(self, task_name, problem = None):
        """
        Constructor
        """
        self.types_present = [] # Data types present in the dataset
        self.task_name = task_name # ML task
        self.solutions = [] # Placeholder to contain list of all pipelines for the specific dataset/problem
        self.basic_sol = None # Primary solution containing all data processing/featurization steps
        self.addn_sol = None # Secondary solution (in case of TEXT/IMAGE etc)
        self.problem = problem
        self.rows = 0
        self.total_cols = 0
        self.metric = None
        self.timeout_in_min = 0

    def add_classification_pipelines(self):
        # Add a classification pipeline too
        primitives = ['d3m.primitives.classification.random_forest.SKlearn',
                      'd3m.primitives.classification.extra_trees.SKlearn',
                      'd3m.primitives.classification.gradient_boosting.SKlearn',
                      'd3m.primitives.classification.bernoulli_naive_bayes.SKlearn']
        for p in primitives:
            pipe = self.get_solution_by_task_name('CLASSIFICATION', p)
            pipe.set_classifier_hyperparam()
            self.solutions.append(pipe)

    def add_regression_pipelines(self):
        # Add a regression pipeline too
        primitives = ['d3m.primitives.regression.random_forest.SKlearn',
                      'd3m.primitives.regression.extra_trees.SKlearn',
                      'd3m.primitives.regression.gradient_boosting.SKlearn']

        for p in primitives:
            pipe = self.get_solution_by_task_name('REGRESSION', p)
            self.solutions.append(pipe)

    def get_link_prediction_timeseries_pipelines(self):
        names = ['LINKPREDICTION','LINKPREDICTION2']
        for name in names:
            pipe = self.get_solution_by_task_name(name)
            self.solutions.append(pipe)
        self.add_regression_pipelines()

    def get_vertex_classification_pipelines(self):
        names = ['VERTEXCLASSIFICATION','VERTEXCLASSIFICATION2','VERTEXCLASSIFICATION3']
        for name in names:
            pipe = self.get_solution_by_task_name(name)
            self.solutions.append(pipe)
        self.add_classification_pipelines()
        prims = ['d3m.primitives.classification.extra_trees.SKlearn', 'd3m.primitives.classification.random_forest.SKlearn']
        for prim in prims:
            pipe = self.get_solution_by_task_name('ISI_PIPELINE', ML_prim=prim, outputstep=2, dataframestep=1)
            self.solutions.append(pipe)

    def get_graph_matching_pipelines(self):
        names = ['GRAPHMATCHING','GRAPHMATCHING2','GRAPHMATCHING3']
        for name in names:
            pipe = self.get_solution_by_task_name(name)
            self.solutions.append(pipe)
        self.add_classification_pipelines()

    def get_link_prediction_pipelines(self):
        names = ['LINKPREDICTION','LINKPREDICTION2']
        for name in names:
            pipe = self.get_solution_by_task_name(name)
            self.solutions.append(pipe)
        self.add_classification_pipelines()

    def get_collaborative_filtering_pipelines(self):
        pipe = self.get_solution_by_task_name(self.task_name)
        self.solutions.append(pipe)

        # Add a regression pipeline too
        pipe = self.get_solution_by_task_name('REGRESSION', 'd3m.primitives.regression.extra_trees.SKlearn')
        self.solutions.append(pipe)

    def get_SSL_pipelines(self, rows):
        basics = [self.basic_sol, self.addn_sol]
        construct_primitive = 'd3m.primitives.data_transformation.construct_predictions.Common'
        if 'ROC_AUC' in self.metric:
            construct_primitive='d3m.primitives.data_transformation.construct_confidence.Common'

        # Iterate through variants of possible blackbox hyperparamets.
        total_cols = self.total_cols
        for variant in solution_templates.sslVariants:
            valid = is_primitive_reasonable(variant, rows, total_cols, self.types_present, self.task_name, self.metric)
            if valid is False:
                continue

            for sol in basics:
                if sol is None:
                    continue

                pipe = copy.deepcopy(sol)
                pipe.id = str(uuid.uuid4())
                pipe.add_step('d3m.primitives.semisupervised_classification.iterative_labeling.AutonBox', outputstep = pipe.index_denormalize + 3, construct_primitive=construct_primitive)
                pipe.add_ssl_variant(variant)
                self.solutions.append(pipe)

        # Add extra pipeline too
        pipe = self.get_solution_by_task_name('SEMISUPERVISED_HDB', 'd3m.primitives.semisupervised_classification.iterative_labeling.AutonBox')
        pipe.add_ssl_variant('d3m.primitives.classification.random_forest.SKlearn')
        self.solutions.append(pipe)

    def get_community_detection_pipelines(self):
        names = ['COMMUNITYDETECTION','COMMUNITYDETECTION2']
        for name in names:
            pipe = self.get_solution_by_task_name(name)
            self.solutions.append(pipe)

    def get_clustering_pipelines(self):
        pipe = self.get_solution_by_task_name(self.task_name)
        self.solutions.append(pipe)
        self.add_classification_pipelines()

    def get_forecasting_pipelines(self, dataset):
        names = ['ARIMA_FORECASTING','VAR_FORECASTING', 'NBEATS_FORECASTING', 'DISTIL_NBEATS', 'ESRNN_FORECASTING']#,'LSTM_FORECASTING']
        for name in names:
            pipe = self.get_solution_by_task_name(name)
            self.solutions.append(pipe)
        self.task_name = 'FORE_REGRESSION'
        self.get_solutions(dataset, self.metric, self.timeout_in_min)
 
    def get_object_detection_pipelines(self):
        names = ['OBJECTDETECTION','OBJECTDETECTION2']
        for name in names:
            pipe = self.get_solution_by_task_name(name)
            self.solutions.append(pipe)

    def create_basic_solutions(self, dataset):
        """
        In case of tasks with multiple pipelines (classification/regression/SSL), primary data processing/featurization steps are
        run only once and cached for all the pipelines.
        """
        if self.task_name != 'CLASSIFICATION' and \
           self.task_name != 'REGRESSION' and \
           self.task_name != 'SEMISUPERVISED' and \
           self.task_name != 'FORE_REGRESSION' and \
           self.task_name != 'REMOTESENSING':
            return

        basic_sol = None
        try:
            # Set data types, and meta data information to begin with
            basic_sol = solutiondescription.SolutionDescription(self.problem)
            basic_sol.initialize_solution(self.task_name)
            (self.types_present, self.total_cols, self.rows, categorical_atts, ordinal_atts, ok_to_denormalize, privileged, add_floats) = helper_functions.column_types_present(dataset)
            logging.critical(self.types_present)
            basic_sol.set_add_floats(add_floats)
            basic_sol.set_categorical_atts(categorical_atts)
            basic_sol.set_ordinal_atts(ordinal_atts)
            basic_sol.set_denormalize(ok_to_denormalize)
            basic_sol.set_privileged(privileged)
            basic_sol.initialize_solution(self.task_name, self.total_cols, self.metric)
        except Exception as e:
            print(e)
            logging.error(sys.exc_info()[0])
            basic_sol = None
            self.types_present = None
            self.basic_sol = None
            return

        # For file in each data point, we treat as time series for featurization
        if len(self.types_present) == 1 and self.types_present[0] == 'FILES':
            self.types_present[0] = 'TIMESERIES'

        self.basic_sol = basic_sol

        if self.task_name == 'REMOTESENSING':
            return

        # Initialize basic solutions based on data types- TIMESERIES / IMAGE / TEXT / AUDIO / VIDEO 
        if 'TIMESERIES' in self.types_present:
            self.basic_sol.initialize_solution('TIMESERIES')
        elif 'IMAGE' in self.types_present:
            if self.rows > default_hyperparams.multimedia_threshold:
                self.basic_sol.add_splitter()
                self.rows = default_hyperparams.multimedia_threshold
            self.basic_sol.initialize_solution('IMAGE', self.total_cols, self.metric)
            self.addn_sol = copy.deepcopy(basic_sol)
            self.addn_sol.initialize_solution('IMAGE2', self.total_cols, self.metric)
        elif 'COREXTEXT' in self.types_present:
            if self.task_name == "FORE_REGRESSION": # Forecasting
                self.basic_sol.initialize_solution('DISTILTEXT', self.total_cols, self.metric)
                self.basic_sol.set_distil_text_hyperparam()
            else: # Classification / Regression
                self.basic_sol.initialize_solution('DISTILTEXT', self.total_cols, self.metric)
                if self.task_name == 'REGRESSION':
                    self.basic_sol.set_distil_text_hyperparam()
                if self.rows < 1000000: # Add Corex pipelines for rows < 1000K
                    self.addn_sol = copy.deepcopy(basic_sol)
                    self.addn_sol.initialize_solution('COREXTEXT')
                if self.rows < 100000: # Add TFIDF pipelines for rows < 100K
                    self.add_TFIDF_solutions()
        elif 'AUDIO' in self.types_present:
            if self.rows > default_hyperparams.multimedia_threshold:
                self.basic_sol.add_splitter()
                self.rows = default_hyperparams.multimedia_threshold
            self.basic_sol.initialize_solution('AUDIO', self.total_cols, self.metric)
        elif 'VIDEO' in self.types_present:
            if self.rows > default_hyperparams.multimedia_threshold:
                self.basic_sol.add_splitter()
                self.rows = default_hyperparams.multimedia_threshold
            self.basic_sol.initialize_solution('VIDEO', self.total_cols, self.metric)

    def add_TFIDF_solutions(self):
        """
        Add solutions for TFIDF text featurization
        """
        basic_sol = solutiondescription.SolutionDescription(self.problem)
        basic_sol.initialize_solution('TFIDFTEXT')
        self.get_primitive_solutions(basic_sol, self.rows)

    def get_solutions(self, dataset, problem_metric, timeout_in_min):
        """
        Get a list of available solutions(pipelines) for the specified task
        Used by both TA2 in "search" phase and TA2-TA3 API.
        """
        self.metric = problem_metric
        self.timeout_in_min = timeout_in_min

        if self.task_name == 'VERTEXNOMINATION' or self.task_name == 'VERTEXCLASSIFICATION':
            self.task_name = 'VERTEXCLASSIFICATION'
            self.get_vertex_classification_pipelines()
        elif self.task_name == 'COMMUNITYDETECTION':
            self.get_community_detection_pipelines()
        elif self.task_name == 'LINKPREDICTION':
            self.get_link_prediction_pipelines()
        elif self.task_name == 'GRAPHMATCHING':
            self.get_graph_matching_pipelines()
        elif self.task_name == 'CLUSTERING':
            self.get_clustering_pipelines()
        elif self.task_name == 'OBJECTDETECTION':
            self.get_object_detection_pipelines()
        elif self.task_name == 'COLLABORATIVEFILTERING':
            self.get_collaborative_filtering_pipelines()
        elif self.task_name == 'FORECASTING':
            self.get_forecasting_pipelines(dataset)
        elif self.task_name == 'LINKPREDICTIONTIMESERIES':
            self.get_link_prediction_timeseries_pipelines()
        else: # CLASSIFICATION / REGRESSION / SEMISUPERVISED / FORE_REGRESSION 
            # Initialize dataset properties, data types etc. 
            self.create_basic_solutions(dataset)
           
            # Run common steps of solutions before forking out processes for classifiers/regressors
            output_step_index = self.basic_sol.index_denormalize + 3
            if self.task_name == 'REMOTESENSING':
                output_step_index = self.basic_sol.index_denormalize + 5
            if 'AUDIO' in self.types_present:
                output_step_index = self.basic_sol.index_denormalize + 2

            # Run basic solution (1-time pass of pipeline steps of data preprocessing and cache final output)
            if self.basic_sol.splitter_present() == False or complex_types_present(self.types_present) == True:
                try:
                    self.basic_sol.run_basic_solution(inputs=[dataset], output_step = output_step_index)
                except Exception as e:
                    print(e)
                    logging.error(sys.exc_info()[0])
                    self.basic_sol = None
                  
            # Run additional(secondary solution), if any
            if self.addn_sol is not None and (self.addn_sol.splitter_present() == False or 'IMAGE' in self.types_present):
                try:
                    self.addn_sol.run_basic_solution(inputs=[dataset], output_step = output_step_index)
                except Exception as e:
                    print(e)
                    logging.error(sys.exc_info()[0])
                    self.addn_sol = None

            # Add primitives to basic solutions
            if self.task_name == 'CLASSIFICATION' or self.task_name == 'REMOTESENSING' or self.task_name == 'REGRESSION' or self.task_name == 'FORE_REGRESSION':
                for sol in [self.basic_sol, self.addn_sol]:
                    self.get_primitive_solutions(sol, self.rows)
                # Try RPI solutions
                if self.basic_sol is not None and \
                    self.basic_sol.splitter == False and \
                    self.is_multi_column_output(self.basic_sol, output_step_index) is False and \
                    'ROC_AUC' not in self.metric:
                        self.get_rpi_solutions(self.basic_sol.add_floats, self.basic_sol.privileged, self.rows, dataset)
                conditioner_prim = 'd3m.primitives.classification.bagging.SKlearn'
                if self.task_name != 'CLASSIFICATION':
                    conditioner_prim = 'd3m.primitives.regression.bagging.SKlearn'
                pipe = self.get_solution_by_task_name('CONDITIONER', ML_prim=conditioner_prim)
                if self.rows < 1000000 and self.total_cols < 1000 and \
                   self.task_name != 'REMOTESENSING' and \
                   complex_types_present(self.types_present) == False and \
                   'ROC_AUC' not in self.metric: #'FORE_REGRESSION': 
                    self.solutions.append(pipe)
            else:
                # Add primitives for SSL solutions
                self.get_SSL_pipelines(self.rows)
                self.get_ssl_rpi_solutions(self.basic_sol.add_floats, self.basic_sol.privileged, self.rows, dataset)

        return self.solutions

    def is_multi_column_output(self, basic_sol, outputstep):
        """
        Does dataset have multiple outputs?
        """
        if basic_sol.primitives_outputs != None and len(basic_sol.primitives_outputs[outputstep].columns) > 1:
            return True
        return False

    def get_primitive_solutions(self, basic_sol, rows):
        """
        Create multiple pipelines for a task by iterating through different primitives
        Iterate through different ML models (classifiers/regressors).
        Each one is evaluated in a seperate process. 
        """
        if basic_sol is None:
            return 

        if self.task_name == "REGRESSION" or self.task_name == "FORE_REGRESSION":
            listOfSolutions = solution_templates.regressors
        else:
            listOfSolutions = solution_templates.classifiers
        
        total_cols = self.total_cols
        dataframestep = 1
        if basic_sol.splitter_present() == True:
            dataframestep = 2

        if complex_types_present(self.types_present) is True and self.task_name == "CLASSIFICATION":
            listOfSolutions.append('d3m.primitives.classification.bernoulli_naive_bayes.SKlearn')

        outputstep = basic_sol.index_denormalize + 3
        if 'AUDIO' in self.types_present:
            outputstep = basic_sol.index_denormalize + 2
        if self.task_name == 'REMOTESENSING':
            outputstep = basic_sol.index_denormalize + 5
            dataframestep = 3
            listOfSolutions = ['d3m.primitives.classification.linear_svc.SKlearn',
                               'd3m.primitives.classification.logistic_regression.SKlearn',
                               'd3m.primitives.classification.random_forest.SKlearn']
      
        construct_primitive = 'd3m.primitives.data_transformation.construct_predictions.Common' 
        if 'ROC_AUC' in self.metric:
            construct_primitive='d3m.primitives.data_transformation.construct_confidence.Common'

        if self.total_cols > 30000:
            if self.task_name == "REGRESSION":
                python_path = 'd3m.primitives.regression.random_forest.SKlearn'
            else:
                python_path = 'd3m.primitives.classification.random_forest.SKlearn'
            basic_sol.add_step(python_path, outputstep = outputstep, dataframestep=dataframestep, construct_primitive=construct_primitive)
            self.solutions.append(basic_sol)
            return

        # Iterate through different primitives (classifiers/regressors)
        for python_path in listOfSolutions:
            # Prune out expensive pipelines
            valid = is_primitive_reasonable(python_path, rows, total_cols, self.types_present, self.task_name, self.metric)
            if valid is False: 
                continue

            if self.is_multi_column_output(basic_sol, outputstep) is True: # multi-column output?
                # These do not work for multi-column output
                if 'linear_sv' in python_path or 'ada_boost' in python_path or 'lasso_cv' in python_path or 'gradient_boosting' in python_path:
                    continue

            # Create new solution by copying basic_sol and appending model.
            # Useful because basic_sol has performed common steps and outputs are cached! 
            pipe = copy.deepcopy(basic_sol)
            pipe.id = str(uuid.uuid4())
            pipe.add_step(python_path, outputstep = outputstep, dataframestep=dataframestep, construct_primitive=construct_primitive)
            self.solutions.append(pipe)

    def get_solution_by_task_name(self, name, ML_prim=None, outputstep=3, dataframestep=1):
        """
        Get a single complete pipeline by task name.
        Used for less frequent tasks (not Classification or regression). 
        name: Name of pipeline template
        ML_prim: ML primitive to be appended (classifier/regressor mostly)
        outputstep: Step in pipeline producing targets
        dataframestep: Dataframe step in pipeline. Used for constructing predictions.
        """
        pipe = solutiondescription.SolutionDescription(self.problem)
        pipe.initialize_solution(name)
        pipe.id = str(uuid.uuid4())
        if ML_prim is not None:
            step1 = pipe.index_denormalize + outputstep
            step2 = pipe.index_denormalize + dataframestep
            pipe.add_step(ML_prim, outputstep=step1, dataframestep=step2)
        else:
            pipe.add_outputs()
        return pipe    

    def get_ssl_rpi_solutions(self, add_floats, privileged, rows, dataset):
        """
        Get RPI-based pipelines for SSL tasks.
        """
        if self.types_present is None:
            return

        if 'AUDIO' in self.types_present or \
           'VIDEO' in self.types_present or \
           'TIMESERIES' in self.types_present or \
           'IMAGE' in self.types_present or \
           'COREXTEXT' in self.types_present:
            return

        basic_sol = solutiondescription.SolutionDescription(self.problem)
        basic_sol.set_privileged(privileged)
        basic_sol.set_add_floats(add_floats)
        basic_sol.initialize_RPI_solution('NOTUNE_PIPELINE_RPI')
        outputstep = basic_sol.index_denormalize + 4
        if add_floats is not None and len(add_floats) > 0:
            outputstep = basic_sol.index_denormalize + 5

        sslVariants = solution_templates.sslVariants
        total_cols = self.total_cols
        if rows <= 100000:
            try:
                basic_sol.run_basic_solution(inputs=[dataset], output_step=outputstep, dataframe_step=basic_sol.index_denormalize + 1)
                total_cols = basic_sol.get_total_cols()
                logging.info("Total cols = %s", total_cols)
            except Exception as e:
                print(e)
                logging.error(sys.exc_info()[0])
                basic_sol = None
        else:
            sslVariants = ['d3m.primitives.classification.random_forest.SKlearn',
                           'd3m.primitives.classification.bagging.SKlearn']

        if basic_sol is None or total_cols > 200:
            return

        # Iterate through different primitives (classifiers)
        for python_path in sslVariants:
            valid = is_primitive_reasonable(python_path, rows, total_cols, self.types_present, self.task_name, self.metric)
            if valid is False:
                continue

            pipe = copy.deepcopy(basic_sol)
            pipe.id = str(uuid.uuid4())
            pipe.add_step('d3m.primitives.semisupervised_classification.iterative_labeling.AutonBox', outputstep = pipe.index_denormalize + 4)
            pipe.add_ssl_variant(python_path)
            self.solutions.append(pipe)

    def get_rpi_solutions(self, add_floats, privileged, rows, dataset):
        """
        Get RPI-based pipelines for classification/regression tasks.
        """
        if self.types_present is None:
            return

        if self.task_name != "REGRESSION" and self.task_name != "CLASSIFICATION":
            return

        if (rows >= 100000 and self.total_cols > 25) or self.total_cols >= 100:
            return

        if 'AUDIO' in self.types_present or \
           'VIDEO' in self.types_present or \
           'TIMESERIES' in self.types_present or \
           'IMAGE' in self.types_present or \
           'COREXTEXT' in self.types_present:
            return

        basic_sol = solutiondescription.SolutionDescription(self.problem)
        basic_sol.set_privileged(privileged)
        basic_sol.set_add_floats(add_floats)
        outputstep = basic_sol.index_denormalize + 4
        if add_floats is not None and len(add_floats) > 0:
            outputstep = basic_sol.index_denormalize + 5

        total_cols = self.total_cols
        if self.task_name == "REGRESSION":
            listOfSolutions = solution_templates.regressors_rpi
        else:
            listOfSolutions = solution_templates.classifiers_rpi

        if rows >= 25000:
            # No tuning
            basic_sol.initialize_RPI_solution('NOTUNE_PIPELINE_RPI')
            if self.task_name == "REGRESSION":
                listOfSolutions = ['d3m.primitives.regression.random_forest.SKlearn']
            else:
                listOfSolutions = ['d3m.primitives.classification.random_forest.SKlearn']
        else:
            # Grid-search over RPI's binsize and model's no. of estimators. This can be expensive
            basic_sol.initialize_RPI_solution('PIPELINE_RPI')
            try:
                basic_sol.run_basic_solution(inputs=[dataset], output_step=outputstep, dataframe_step=basic_sol.index_denormalize + 1)
                total_cols = basic_sol.get_total_cols()
                logging.info("Total cols = %s", total_cols)
            except Exception as e:
                print(e)
                logging.error(sys.exc_info()[0])
                basic_sol = None

        if basic_sol is None or total_cols > 200:
            return

        RPI_steps = ['d3m.primitives.feature_selection.joint_mutual_information.AutoRPI','d3m.primitives.feature_selection.simultaneous_markov_blanket.AutoRPI']

        # Iterate through different primitives (classifiers/regressors)
        for python_path in listOfSolutions:
            # Avoid expensive solutions!!!
            if 'gradient_boosting' in python_path and ((rows > 1000 and total_cols > 50) or (rows > 5000) or total_cols > 100):
                continue

            if rows >= 25000:
                pipe = copy.deepcopy(basic_sol)
                pipe.id = str(uuid.uuid4())
                pipe.add_step(python_path, outputstep)
                self.solutions.append(pipe)
            else:
                for step in RPI_steps:
                    pipe = copy.deepcopy(basic_sol)
                    pipe.id = str(uuid.uuid4())
                    pipe.add_RPI_step(step, python_path, outputstep)
                    self.solutions.append(pipe)

